description = [
    {
        "description": "Analyze cell migration metrics from time-lapse microscopy images.",
        "name": "analyze_cell_migration_metrics",
        "optional_parameters": [
            {
                "default": 1.0,
                "description": "Conversion factor from pixels to micrometers",
                "name": "pixel_size_um",
                "type": "float",
            },
            {
                "default": 1.0,
                "description": "Time interval between consecutive frames in minutes",
                "name": "time_interval_min",
                "type": "float",
            },
            {
                "default": 10,
                "description": "Minimum number of frames a cell must be tracked to be included in analysis",
                "name": "min_track_length",
                "type": "int",
            },
            {
                "default": "./",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the directory containing time-lapse images or path to a multi-frame TIFF file",
                "name": "image_sequence_path",
                "type": "str",
            }
        ],
    },
    {
        "description": "Simulates CRISPR-Cas9 genome editing process including guide "
        "RNA design, delivery, and analysis.",
        "name": "perform_crispr_cas9_genome_editing",
        "optional_parameters": [],
        "required_parameters": [
            {
                "default": None,
                "description": "List of guide RNA sequences (20 "
                "nucleotides each) targeting the "
                "genomic region of interest",
                "name": "guide_rna_sequences",
                "type": "List[str]",
            },
            {
                "default": None,
                "description": "Target genomic sequence to be "
                "edited (should be longer than guide "
                "RNA and contain the target sites)",
                "name": "target_genomic_loci",
                "type": "str",
            },
            {
                "default": None,
                "description": "Type of cell or tissue being edited (affects delivery efficiency and editing outcomes)",
                "name": "cell_tissue_type",
                "type": "str",
            },
        ],
    },
    {
        "description": "Analyze calcium imaging data to quantify neuronal activity "
        "metrics including cell counts, event rates, decay times, and "
        "signal-to-noise ratios.",
        "name": "analyze_calcium_imaging_data",
        "optional_parameters": [
            {
                "default": "./",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            }
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the time-series stack of fluorescence microscopy images (TIFF format)",
                "name": "image_stack_path",
                "type": "str",
            }
        ],
    },
    {
        "description": "Analyzes in vitro drug release kinetics from biomaterial formulations.",
        "name": "analyze_in_vitro_drug_release_kinetics",
        "optional_parameters": [
            {
                "default": "Drug",
                "description": "Name of the drug being analyzed",
                "name": "drug_name",
                "type": "str",
            },
            {
                "default": None,
                "description": "Total amount of drug initially "
                "loaded in the formulation. If None, "
                "the maximum concentration is used "
                "as 100%",
                "name": "total_drug_loaded",
                "type": "float",
            },
            {
                "default": "./",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Time points at which drug concentrations were measured (in hours)",
                "name": "time_points",
                "type": "List[float] or numpy.ndarray",
            },
            {
                "default": None,
                "description": "Measured drug concentration at each time point",
                "name": "concentration_data",
                "type": "List[float] or numpy.ndarray",
            },
        ],
    },
    {
        "description": "Quantifies morphological properties of myofibers in microscopy images of tissue sections.",
        "name": "analyze_myofiber_morphology",
        "optional_parameters": [
            {
                "default": 2,
                "description": "Channel index containing nuclei staining (DAPI, Hoechst, etc.)",
                "name": "nuclei_channel",
                "type": "int",
            },
            {
                "default": 1,
                "description": "Channel index containing myofiber staining (α-Actinin, etc.)",
                "name": "myofiber_channel",
                "type": "int",
            },
            {
                "default": "otsu",
                "description": "Method for thresholding ('otsu', 'adaptive', or 'manual')",
                "name": "threshold_method",
                "type": "str",
            },
            {
                "default": "./",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the microscopy image file "
                "(typically a multichannel image "
                "with nuclei and myofiber staining)",
                "name": "image_path",
                "type": "str",
            }
        ],
    },
    {
        "description": "Model neural activity trajectories and decode behavioral variables.",
        "name": "decode_behavior_from_neural_trajectories",
        "optional_parameters": [
            {
                "default": 10,
                "description": "Number of principal components to use for dimensionality reduction",
                "name": "n_components",
                "type": "int",
            },
            {
                "default": "./",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Neural spiking activity data, shape (n_timepoints, n_neurons)",
                "name": "neural_data",
                "type": "numpy.ndarray",
            },
            {
                "default": None,
                "description": "Behavioral data, shape (n_timepoints, n_behavioral_variables)",
                "name": "behavioral_data",
                "type": "numpy.ndarray",
            },
        ],
    },
    {
        "description": "Simulate a whole-cell model represented as a system of ordinary differential equations (ODEs).",
        "name": "simulate_whole_cell_ode_model",
        "optional_parameters": [
            {
                "default": None,
                "description": "Function defining the system of "
                "ODEs. Should take arguments (t, y, "
                "*args) where t is time, y is the "
                "state vector, and args contains "
                "additional parameters. If None, a "
                "simple example whole-cell model "
                "will be used.",
                "name": "ode_function",
                "type": "callable",
            },
            {
                "default": "(0, 100)",
                "description": "Tuple of (start_time, end_time) for the simulation.",
                "name": "time_span",
                "type": "tuple",
            },
            {
                "default": 1000,
                "description": "Number of time points to evaluate.",
                "name": "time_points",
                "type": "int",
            },
            {
                "default": "'LSODA'",
                "description": "Numerical integration method to use (e.g., 'RK45', 'LSODA', 'BDF').",
                "name": "method",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Initial values for each state "
                "variable in the model. If dict, "
                "keys are variable names and values "
                "are initial concentrations/values. "
                "If array-like, order must match the "
                "order expected by the ODE function.",
                "name": "initial_conditions",
                "type": "dict or array-like",
            },
            {
                "default": None,
                "description": "Model parameters required by the "
                "ODE function. Keys are parameter "
                "names and values are parameter "
                "values.",
                "name": "parameters",
                "type": "dict",
            },
        ],
    },
]



description = [
    {
        "description": "Engineer a bacterial genome by integrating therapeutic genetic parts for therapeutic delivery.",
        "name": "engineer_bacterial_genome_for_therapeutic_delivery",
        "optional_parameters": [],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the file containing the bacterial genome sequence in FASTA format",
                "name": "bacterial_genome_file",
                "type": "str",
            },
            {
                "default": None,
                "description": "Dictionary containing genetic parts "
                "to be integrated (promoters, genes, "
                "terminators, cargo)",
                "name": "genetic_parts",
                "type": "dict",
            },
        ],
    },
    {
        "description": "Analyze bacterial growth data and extract growth parameters from OD600 measurements.",
        "name": "analyze_bacterial_growth_rate",
        "optional_parameters": [
            {
                "default": "Unknown strain",
                "description": "Name of the bacterial strain being analyzed",
                "name": "strain_name",
                "type": "str",
            },
            {
                "default": "./",
                "description": "Directory where to save the output files",
                "name": "output_dir",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Time points at which OD600 measurements were taken (in hours)",
                "name": "time_points",
                "type": "List or numpy.ndarray",
            },
            {
                "default": None,
                "description": "Optical density (OD600) measurements corresponding to each time point",
                "name": "od_measurements",
                "type": "List or numpy.ndarray",
            },
        ],
    },
    {
        "description": "Analyze sequencing data to extract, quantify and determine lineage relationships of barcodes.",
        "name": "analyze_barcode_sequencing_data",
        "optional_parameters": [
            {
                "default": None,
                "description": "Regular expression pattern to identify barcodes. If None, will use flanking sequences",
                "name": "barcode_pattern",
                "type": "str",
            },
            {
                "default": None,
                "description": "5' flanking sequence of the barcode region",
                "name": "flanking_seq_5prime",
                "type": "str",
            },
            {
                "default": None,
                "description": "3' flanking sequence of the barcode region",
                "name": "flanking_seq_3prime",
                "type": "str",
            },
            {
                "default": 5,
                "description": "Minimum count threshold for considering a barcode",
                "name": "min_count",
                "type": "int",
            },
            {
                "default": "./results",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the input sequencing file in FASTQ or FASTA format",
                "name": "input_file",
                "type": "str",
            }
        ],
    },
    {
        "description": "Performs bifurcation analysis on a dynamical system and generates a bifurcation diagram.",
        "name": "analyze_bifurcation_diagram",
        "optional_parameters": [
            {
                "default": "Dynamical System",
                "description": "Name of the dynamical system being analyzed, used for plot titles.",
                "name": "system_name",
                "type": "str",
            },
            {
                "default": "./",
                "description": "Directory to save the output files.",
                "name": "output_dir",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "A 2D array where each row "
                "represents a time series for a "
                "specific parameter value. Shape "
                "should be (n_parameter_values, "
                "n_time_points).",
                "name": "time_series_data",
                "type": "numpy.ndarray",
            },
            {
                "default": None,
                "description": "1D array of parameter values "
                "corresponding to each time series. "
                "Shape should be "
                "(n_parameter_values,).",
                "name": "parameter_values",
                "type": "numpy.ndarray",
            },
        ],
    },
    {
        "description": "Generate a mathematical model of a biochemical network in SBML format.",
        "name": "create_biochemical_network_sbml_model",
        "optional_parameters": [
            {
                "default": "biochemical_model.xml",
                "description": "File path to save the SBML model",
                "name": "output_file",
                "type": "str",
            }
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "List of dictionaries representing "
                "reactions with id, name, reactants, "
                "products, and reversible properties",
                "name": "reaction_network",
                "type": "List[dict]",
            },
            {
                "default": None,
                "description": "Dictionary mapping reaction IDs to kinetic law parameters with law_type and parameters",
                "name": "kinetic_parameters",
                "type": "dict",
            },
        ],
    },
    {
        "description": "Analyzes and optimizes a DNA/RNA sequence for improved "
        "expression in a heterologous host organism.",
        "name": "optimize_codons_for_heterologous_expression",
        "optional_parameters": [],
        "required_parameters": [
            {
                "default": None,
                "description": "The DNA or RNA sequence of the "
                "target gene to be optimized. Should "
                "contain complete codons (length "
                "divisible by 3).",
                "name": "target_sequence",
                "type": "str",
            },
            {
                "default": None,
                "description": "Dictionary mapping codons to their "
                "usage frequency in the host "
                "organism. Format: {'AUG': 0.8, "
                "'GCC': 0.6, ...} or {'ATG': 0.8, "
                "'GCC': 0.6, ...}",
                "name": "host_codon_usage",
                "type": "dict",
            },
        ],
    },
    {
        "description": "Simulate gene regulatory circuit dynamics with growth feedback.",
        "name": "simulate_gene_circuit_with_growth_feedback",
        "optional_parameters": [
            {
                "default": 100,
                "description": "Total simulation time",
                "name": "simulation_time",
                "type": "float",
            },
            {
                "default": 1000,
                "description": "Number of time points to sample",
                "name": "time_points",
                "type": "int",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Adjacency matrix representing the "
                "gene circuit topology. Positive "
                "values indicate activation, "
                "negative values indicate "
                "repression. Shape should be "
                "(n_genes, n_genes) where n_genes is "
                "the number of genes in the circuit.",
                "name": "circuit_topology",
                "type": "numpy.ndarray",
            },
            {
                "default": None,
                "description": "Dictionary containing kinetic "
                "parameters: 'basal_rates', "
                "'degradation_rates', "
                "'hill_coefficients', and "
                "'threshold_constants' for the gene "
                "circuit.",
                "name": "kinetic_params",
                "type": "dict",
            },
            {
                "default": None,
                "description": "Dictionary containing "
                "growth-related parameters: "
                "'max_growth_rate', "
                "'growth_inhibition', and "
                "'gene_growth_weights'.",
                "name": "growth_params",
                "type": "dict",
            },
        ],
    },
    {
        "description": "Identifies functional domains within a Fatty Acid Synthase "
        "(FAS) sequence and predicts their roles.",
        "name": "identify_fas_functional_domains",
        "optional_parameters": [
            {
                "default": "protein",
                "description": 'Type of sequence provided - "protein" or "nucleotide"',
                "name": "sequence_type",
                "type": "str",
            },
            {
                "default": "fas_domains_report.txt",
                "description": "Name of the output file to save the detailed domain report",
                "name": "output_file",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "The nucleotide or protein sequence of a FAS gene",
                "name": "sequence",
                "type": "str",
            }
        ],
    },
]



#!/usr/bin/env python3
"""Command-line tool to generate Python functions from task descriptions using the function_generator agent."""

import argparse
import os
import re
import sys
from pathlib import Path

# Add the parent directory to the path so we can import bioagentos
sys.path.append(str(Path(__file__).resolve().parent.parent.parent))

from bioagentos.agent.function_generator import function_generator


def sanitize_filename(name):
    """Convert a string to a valid filename by replacing invalid characters.

    Args:
        name: The string to convert

    Returns:
        A sanitized filename

    """
    # Replace spaces and special characters with underscores
    sanitized = re.sub(r"[^\w\s-]", "_", name.lower())
    sanitized = re.sub(r"[\s-]+", "_", sanitized)
    return sanitized


def main():
    """Main function for the command-line tool."""
    parser = argparse.ArgumentParser(description="Generate a Python function from a task description")
    parser.add_argument(
        "--task",
        "-t",
        type=str,
        help="Task description (if not provided, will prompt for input)",
    )
    parser.add_argument(
        "--output-dir",
        "-o",
        type=str,
        default="generated_functions",
        help="Directory to save the generated function (default: generated_functions)",
    )
    parser.add_argument(
        "--filename",
        "-f",
        type=str,
        help="Base filename for the generated function (without extension, default: derived from function name)",
    )
    parser.add_argument(
        "--model",
        "-m",
        type=str,
        default="claude-3-7-sonnet-latest",
        help="LLM model to use (default: claude-3-7-sonnet-latest)",
    )
    parser.add_argument(
        "--temperature",
        type=float,
        default=0.7,
        help="Temperature setting for the LLM (default: 0.7)",
    )

    args = parser.parse_args()

    # If no task is provided, prompt for it
    if not args.task:
        print("Please enter your task description (type 'END' on a new line when finished):")
        lines = []
        while True:
            line = input()
            if line.strip() == "END":
                break
            lines.append(line)
        task_description = "\n".join(lines)
    else:
        task_description = args.task

    # Create the output directory if it doesn't exist
    os.makedirs(args.output_dir, exist_ok=True)

    # Initialize the function generator agent
    agent = function_generator(llm=args.model, temperature=args.temperature)

    # Generate the function
    print(f"\nGenerating function using {args.model}...")
    result = agent.generate_function(task_description)

    # Extract the function name from the generated code
    function_name = "generated_function"  # Default name
    function_code = result["function_code"]
    if function_code:
        match = re.search(r"def\s+(\w+)", function_code)
        if match:
            function_name = match.group(1)

    # Determine the filename
    base_filename = sanitize_filename(args.filename) if args.filename else sanitize_filename(function_name)

    # Save the function to a file
    function_file = f"{args.output_dir}/{base_filename}.py"
    test_file = f"{args.output_dir}/test_{base_filename}.py"

    with open(function_file, "w") as f:
        f.write(function_code)

    with open(test_file, "w") as f:
        f.write(result["test_case"])

    # Print the results
    print("\n" + "=" * 80)
    print("GENERATED FUNCTION:")
    print("=" * 80)
    print(function_code)

    print("\n" + "=" * 80)
    print("TEST CASE:")
    print("=" * 80)
    print(result["test_case"])

    print(f"\nFunction saved to {function_file}")
    print(f"Test case saved to {test_file}")


if __name__ == "__main__":
    main()



description = [
    {
        "description": "Executes the provided Python command in the notebook environment and returns the output.",
        "name": "run_python_repl",
        "optional_parameters": [],
        "required_parameters": [
            {
                "default": None,
                "description": "Python command to execute in the notebook environment",
                "name": "command",
                "type": "str",
            }
        ],
    },
    {
        "description": "Read the source code of a function from any module path.",
        "name": "read_function_source_code",
        "optional_parameters": [],
        "required_parameters": [
            {
                "default": None,
                "description": "Fully qualified function name "
                "(e.g., "
                "'bioagentos.tool.support_tools.write_python_code')",
                "name": "function_name",
                "type": "str",
            }
        ],
    },
]



#!/usr/bin/env python3
"""Script to extract common tasks, databases, and software from bioRxiv papers
using the PaperTaskExtractor and existing metadata CSV.
"""

import argparse
import io
import json
import os
import random
import sys
import time
from typing import Any

import pandas as pd
import PyPDF2
import requests
from tqdm import tqdm

# Add the parent directory to the path so we can import the bioagentos package
sys.path.append("../../")
from biomni.agent.env_collection import PaperTaskExtractor


def parse_arguments():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description="Extract tasks from bioRxiv papers in a specific subject area.")
    parser.add_argument(
        "--subject",
        type=str,
        required=True,
        help='Subject area to search for papers (e.g., "neuroscience", "bioinformatics")',
    )
    parser.add_argument(
        "--limit",
        type=int,
        default=10,
        help="Maximum number of papers to process (default: 10)",
    )
    parser.add_argument(
        "--metadata-path",
        type=str,
        default="/dfs/user/kexinh/BioAgentOS/data/biorxiv_metadata.csv",
        help="Path to bioRxiv metadata CSV file (default: data/biorxiv_metadata.csv)",
    )
    parser.add_argument(
        "--output-dir",
        type=str,
        default=None,
        help="Directory to save results (default: ./biorxiv_results_{subject}_{limit})",
    )
    parser.add_argument(
        "--model",
        type=str,
        default="claude-3-haiku-20240307",
        help="LLM model to use for extraction (default: claude-3-haiku-20240307)",
    )
    parser.add_argument(
        "--chunk-size",
        type=int,
        default=4000,
        help="Chunk size for text processing (default: 4000)",
    )
    parser.add_argument(
        "--chunk-overlap",
        type=int,
        default=400,
        help="Chunk overlap for text processing (default: 400)",
    )
    parser.add_argument(
        "--max-paper-length",
        type=int,
        default=200000,
        help="Maximum paper length in characters (default: 200000)",
    )
    parser.add_argument("--save-pdfs", action="store_true", help="Save downloaded PDFs (default: False)")
    parser.add_argument(
        "--random-sample",
        action="store_true",
        help="Randomly sample papers instead of taking the first N (default: False)",
    )
    return parser.parse_args()


def load_papers_from_csv(metadata_path: str, subject: str, limit: int, random_sample: bool) -> pd.DataFrame:
    """Load papers from bioRxiv metadata CSV file.

    Args:
        metadata_path: Path to the metadata CSV file
        subject: Subject area to filter by
        limit: Maximum number of papers to return
        random_sample: Whether to randomly sample papers

    Returns:
        DataFrame of filtered papers

    """
    try:
        # Load the metadata CSV
        df_biorxiv = pd.read_csv(metadata_path)

        # Filter published papers in the specified subject
        if subject.lower() == "all":
            filtered_df = df_biorxiv[df_biorxiv.published != "NA"]
        else:
            filtered_df = df_biorxiv[
                (df_biorxiv.published != "NA") & (df_biorxiv.category.str.lower() == subject.lower())
            ]

        # Sample papers
        if len(filtered_df) > limit:
            filtered_df = filtered_df.sample(limit, random_state=42) if random_sample else filtered_df.head(limit)

        print(f"Loaded {len(filtered_df)} papers in subject: {subject}")
        return filtered_df

    except Exception as e:
        print(f"Error loading papers from CSV: {e}")
        return pd.DataFrame()


def download_pdf(paper_doi: str, save_path: str | None = None) -> str | None:
    """Download PDF for a given paper DOI.

    Args:
        paper_doi: DOI of the paper
        save_path: Path to save the PDF (if None, won't save to disk)

    Returns:
        PDF text content or None if download failed

    """
    # Construct PDF URL from DOI
    pdf_url = f"https://www.biorxiv.org/content/{paper_doi}v1.full.pdf"

    try:
        # Add random delay to avoid rate limiting
        time.sleep(random.uniform(1, 3))

        response = requests.get(pdf_url, stream=True)
        response.raise_for_status()

        # Save PDF if requested
        if save_path:
            os.makedirs(os.path.dirname(save_path), exist_ok=True)
            with open(save_path, "wb") as f:
                for chunk in response.iter_content(chunk_size=8192):
                    f.write(chunk)

        # Extract text from PDF
        pdf_text = extract_text_from_pdf(io.BytesIO(response.content))
        return pdf_text

    except requests.exceptions.RequestException as e:
        print(f"Error downloading PDF for {paper_doi}: {e}")
        return None


def extract_text_from_pdf(pdf_file) -> str:
    """Extract text content from a PDF file.

    Args:
        pdf_file: File-like object containing PDF data

    Returns:
        Extracted text content

    """
    try:
        pdf_reader = PyPDF2.PdfReader(pdf_file)
        text = ""
        for page_num in range(len(pdf_reader.pages)):
            page = pdf_reader.pages[page_num]
            text += page.extract_text() + "\n\n"
        return text
    except Exception as e:
        print(f"Error extracting text from PDF: {e}")
        return ""


def truncate_text(text: str, max_length: int) -> str:
    """Truncate text to a maximum length while preserving complete sentences.

    Args:
        text: The text to truncate
        max_length: Maximum length in characters

    Returns:
        Truncated text

    """
    if len(text) <= max_length:
        return text

    # Find the last sentence boundary before max_length
    truncated = text[:max_length]
    last_period = truncated.rfind(".")
    last_question = truncated.rfind("?")
    last_exclamation = truncated.rfind("!")

    # Find the last sentence boundary
    last_boundary = max(last_period, last_question, last_exclamation)

    if last_boundary > 0:
        return text[: last_boundary + 1]
    else:
        # If no sentence boundary found, just truncate at max_length
        return truncated


def process_papers(papers_df: pd.DataFrame, args) -> list[dict[str, Any]]:
    """Process papers using PaperTaskExtractor.

    Args:
        papers_df: DataFrame of paper metadata
        args: Command line arguments

    Returns:
        List of papers with extracted tasks, databases, and software

    """
    # Initialize the task extractor
    extractor = PaperTaskExtractor(llm=args.model, chunk_size=args.chunk_size, chunk_overlap=args.chunk_overlap)

    results = []

    for _, paper in tqdm(papers_df.iterrows(), total=len(papers_df), desc="Processing papers"):
        paper_doi = paper.get("doi")
        if not paper_doi:
            continue

        print(f"\nProcessing paper: {paper.get('title')} (DOI: {paper_doi})")

        # Create paths
        pdf_path = None
        if args.save_pdfs:
            pdf_path = os.path.join(args.output_dir, "pdfs", f"{paper_doi.replace('/', '_')}.pdf")

        result_path = os.path.join(args.output_dir, "results", f"{paper_doi.replace('/', '_')}.json")

        # Skip if already processed
        if os.path.exists(result_path):
            print(f"Paper already processed, loading from {result_path}")
            with open(result_path) as f:
                paper_result = json.load(f)
                results.append(paper_result)
            continue

        # Download and process PDF
        pdf_text = download_pdf(paper_doi, pdf_path)
        if not pdf_text:
            print(f"Failed to download or extract text from {paper_doi}")
            continue

        # Truncate text if it exceeds max length
        if len(pdf_text) > args.max_paper_length:
            original_length = len(pdf_text)
            pdf_text = truncate_text(pdf_text, args.max_paper_length)
            print(f"Truncated paper from {original_length} to {len(pdf_text)} characters")

        # Extract tasks, databases, and software
        try:
            _, extraction_results = extractor.go(pdf_text)

            # Add paper metadata to results
            paper_result = {
                "metadata": {
                    "doi": paper.get("doi"),
                    "title": paper.get("title"),
                    "authors": paper.get("authors", ""),
                    "abstract": paper.get("abstract", ""),
                    "date": paper.get("date", ""),
                    "category": paper.get("category", ""),
                    "text_length": len(pdf_text),
                },
                "extraction": extraction_results,
            }

            # Save results
            os.makedirs(os.path.dirname(result_path), exist_ok=True)
            with open(result_path, "w") as f:
                json.dump(paper_result, f, indent=2)

            results.append(paper_result)

        except Exception as e:
            print(f"Error processing paper {paper_doi}: {e}")

    return results


def generate_summary(results: list[dict[str, Any]], output_dir: str):
    """Generate summary of extracted tasks, databases, and software.

    Args:
        results: List of papers with extraction results
        output_dir: Directory to save summary

    """
    # Collect all tasks, databases, and software
    all_tasks = []
    all_databases = []
    all_software = []

    for paper in results:
        extraction = paper.get("extraction", {})

        # Add paper metadata to each item
        paper_meta = paper.get("metadata", {})
        paper_id = f"{paper_meta.get('title')} ({paper_meta.get('doi')})"

        # Process tasks
        for task in extraction.get("tasks", []):
            task["paper"] = paper_id
            all_tasks.append(task)

        # Process databases
        for db in extraction.get("databases", []):
            db["paper"] = paper_id
            all_databases.append(db)

        # Process software
        for sw in extraction.get("software", []):
            sw["paper"] = paper_id
            all_software.append(sw)

    # Create summary dataframes
    if all_tasks:
        tasks_df = pd.DataFrame(all_tasks)
        tasks_df.to_csv(os.path.join(output_dir, "tasks_summary.csv"), index=False)

    if all_databases:
        db_df = pd.DataFrame(all_databases)
        db_df.to_csv(os.path.join(output_dir, "databases_summary.csv"), index=False)

    if all_software:
        sw_df = pd.DataFrame(all_software)
        sw_df.to_csv(os.path.join(output_dir, "software_summary.csv"), index=False)

    # Generate frequency counts
    task_counts = {}
    for task in all_tasks:
        task_name = task.get("task_name")
        if task_name:
            task_counts[task_name] = task_counts.get(task_name, 0) + 1

    db_counts = {}
    for db in all_databases:
        db_name = db.get("name")
        if db_name:
            db_counts[db_name] = db_counts.get(db_name, 0) + 1

    sw_counts = {}
    for sw in all_software:
        sw_name = sw.get("name")
        if sw_name:
            sw_counts[sw_name] = sw_counts.get(sw_name, 0) + 1

    # Save frequency counts
    with open(os.path.join(output_dir, "frequency_summary.json"), "w") as f:
        json.dump(
            {"tasks": task_counts, "databases": db_counts, "software": sw_counts},
            f,
            indent=2,
        )

    print(f"Summary files saved to {output_dir}")


def main():
    """Main function to run the script."""
    args = parse_arguments()

    # Set default output directory if not specified
    if args.output_dir is None:
        # Clean subject name for directory name (replace spaces with underscores, lowercase)
        clean_subject = args.subject.lower().replace(" ", "_").replace("/", "_")
        args.output_dir = f"./biorxiv_results_{clean_subject}_{args.limit}"

    # Create output directories
    os.makedirs(args.output_dir, exist_ok=True)
    os.makedirs(os.path.join(args.output_dir, "results"), exist_ok=True)
    if args.save_pdfs:
        os.makedirs(os.path.join(args.output_dir, "pdfs"), exist_ok=True)

    # Load papers from CSV
    print(f"Loading papers from {args.metadata_path} in subject area: {args.subject}")
    papers_df = load_papers_from_csv(args.metadata_path, args.subject, args.limit, args.random_sample)

    if len(papers_df) == 0:
        print(f"No papers found for subject: {args.subject}")
        return

    # Process papers
    results = process_papers(papers_df, args)

    # Generate summary
    generate_summary(results, args.output_dir)

    print("Done!")


if __name__ == "__main__":
    main()



import numpy as np
import pandas as pd
from bioagentos.task.base_task import base_task

np.random.seed(42)


def shuffle(x):
    np.random.shuffle(x)
    return x


class lab_bench(base_task):
    def __init__(self, path="./data", dataset="DbQA"):
        if dataset not in ["DbQA", "SeqQA"]:
            raise ValueError("dataset must be one of 'DbQA', 'SeqQA'")

        self.dataset = dataset  # Store dataset type
        df = pd.read_parquet(path + "/" + dataset + "/train-00000-of-00001_test.parquet")

        self.prompt = """The following is a multiple choice question about biology.
Please answer by responding with the letter of the correct answer.

Question: {question}
Options:
{options}

You MUST include the letter of the correct answer within the following tags:
[ANSWER] and [/ANSWER]. For example, '[ANSWER]<answer>[/ANSWER]',
where <answer> is the correct letter. Always answer in exactly this format
of a single letter between the two tags, even if you are unsure.
We require this because we use automatic parsing.
            """

        np.random.seed(42)
        df["options"] = df.apply(
            lambda x: shuffle(
                x.distractors.tolist() + [x.ideal] + ["Insufficient information to answer the question."]
            ),
            axis=1,
        )
        df["options_letters"] = df.options.apply(
            lambda x: "\n".join([chr(ord("A") + i) + "." + item for i, item in enumerate(x)])
        )
        df["letter_answer"] = df.apply(
            lambda x: chr(ord("A") + np.where(np.array(x.options) == x.ideal)[0][0]),
            axis=1,
        )
        df["letter_refrain"] = df.apply(
            lambda x: chr(
                ord("A") + np.where(np.array(x.options) == "Insufficient information to answer the question.")[0][0]
            ),
            axis=1,
        )

        self.query = df.question.values
        self.options = df.options_letters.values
        self.answer = df.letter_answer.values
        self.refrain_label = df.letter_refrain.values

        # Store protocol information if available
        self.protocol = df.protocol.values if "protocol" in df.columns else None

    def get_example(self, index=None):
        if index is None:
            index = np.random.randint(len(self.query))

        if self.dataset == "ProtocolQA" and self.protocol is not None:
            return {
                "prompt": self.prompt.format(
                    protocol=self.protocol[index],
                    question=self.query[index],
                    options=self.options[index],
                ),
                "answer": self.answer[index],
            }
        else:
            return {
                "prompt": self.prompt.format(question=self.query[index], options=self.options[index]),
                "answer": self.answer[index],
            }

    def get_iterator(self):
        for i in range(len(self.query)):
            yield self.get_example(i)

    def evaluate(self, response):
        ## expected a list/array of symbols
        from sklearn.metrics import accuracy_score

        ground_truth = self.answer
        response = np.array(response)

        return {
            "accuracy": accuracy_score(ground_truth, response),
            "coverage": np.mean(response != self.refrain_label),
            "refrain_ratio": np.mean(response == self.refrain_label),
            "precision": accuracy_score(
                ground_truth[np.where(response != self.refrain_label)],
                response[np.where(response != self.refrain_label)],
            ),
        }

    def output_class(self):
        from typing import Optional

        from pydantic import BaseModel, Field

        class MultipleChoiceOutput(BaseModel):
            """Multiple choice output."""

            choice: str | None = Field(
                description="Multiple choice answer. For example, if there is <answer>A</answer> in the prompt, the output should be 'A'."
            )

        return MultipleChoiceOutput



import os
from typing import Optional

from Bio import Entrez
from mcp.server.fastmcp import FastMCP
from pydantic import BaseModel

Entrez.email = os.getenv("NCBI_EMAIL") or "YOUR_EMAIL"
mcp = FastMCP("pubmed-live")


class Article(BaseModel):
    pmid: str
    title: str | None = None
    journal: str | None = None
    year: str | None = None
    abstract: str | None = None


@mcp.tool()
def search_pubmed(query: str, max_results: int = 10) -> list[Article]:
    ids = Entrez.read(Entrez.esearch(db="pubmed", term=query, retmax=max_results))["IdList"]
    summaries = Entrez.read(Entrez.esummary(db="pubmed", id=",".join(ids))) if ids else []
    articles: list[Article] = []
    for s in summaries:
        articles.append(
            Article(pmid=s["Id"], title=s["Title"], journal=s.get("FullJournalName", ""), year=s.get("PubDate", "")[:4])
        )
    return articles


@mcp.tool()
def get_article_abstract(pmid: str) -> str:
    with Entrez.efetch(db="pubmed", id=pmid, rettype="abstract", retmode="text") as h:
        return h.read()


if __name__ == "__main__":
    mcp.run()



description = [
    {
        "description": "Perform liftover of genomic coordinates between hg19 and "
        "hg38 formats with detailed intermediate steps.",
        "name": "liftover_coordinates",
        "optional_parameters": [],
        "required_parameters": [
            {
                "default": None,
                "description": "Chromosome number (e.g., '1', 'X')",
                "name": "chromosome",
                "type": "str",
            },
            {
                "default": None,
                "description": "Genomic position",
                "name": "position",
                "type": "int",
            },
            {
                "default": None,
                "description": "Input genome build ('hg19' or 'hg38')",
                "name": "input_format",
                "type": "str",
            },
            {
                "default": None,
                "description": "Output genome build ('hg19' or 'hg38')",
                "name": "output_format",
                "type": "str",
            },
            {
                "default": None,
                "description": "Path to liftover chain files",
                "name": "data_path",
                "type": "str",
            },
        ],
    },
    {
        "description": "Performs Bayesian fine-mapping from GWAS summary statistics "
        "using deep variational inference to compute posterior "
        "inclusion probabilities and credible sets for putative "
        "causal variants.",
        "name": "bayesian_finemapping_with_deep_vi",
        "optional_parameters": [
            {
                "default": 5000,
                "description": "Number of training iterations for the variational inference algorithm",
                "name": "n_iterations",
                "type": "int",
            },
            {
                "default": 0.01,
                "description": "Learning rate for the optimization algorithm",
                "name": "learning_rate",
                "type": "float",
            },
            {
                "default": 64,
                "description": "Hidden dimension size for the neural network",
                "name": "hidden_dim",
                "type": "int",
            },
            {
                "default": 0.95,
                "description": "Threshold for defining the credible set (e.g., 0.95 for a 95% credible set)",
                "name": "credible_threshold",
                "type": "float",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to CSV or TSV file containing "
                "GWAS summary statistics with "
                "variant_id, effect_size, pvalue, "
                "and optional se columns",
                "name": "gwas_summary_path",
                "type": "str",
            },
            {
                "default": None,
                "description": "Linkage disequilibrium matrix with pairwise correlations between variants",
                "name": "ld_matrix",
                "type": "numpy.ndarray",
            },
        ],
    },
    {
        "description": "Analyzes and categorizes mutations induced by Cas9 at target sites.",
        "name": "analyze_cas9_mutation_outcomes",
        "optional_parameters": [
            {
                "default": None,
                "description": "Dictionary mapping sequence IDs to "
                "cell line information (e.g., "
                "wildtype, knockout gene)",
                "name": "cell_line_info",
                "type": "dict",
            },
            {
                "default": "cas9_mutation_analysis",
                "description": "Prefix for output files",
                "name": "output_prefix",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Dictionary mapping sequence IDs to reference DNA sequences (strings)",
                "name": "reference_sequences",
                "type": "dict",
            },
            {
                "default": None,
                "description": "Nested dictionary: {sequence_id: "
                "{read_id: sequence}} containing the "
                "edited/mutated sequences for each "
                "reference",
                "name": "edited_sequences",
                "type": "dict",
            },
        ],
    },
    {
        "description": "Analyzes CRISPR-Cas9 genome editing results by comparing original and edited sequences.",
        "name": "analyze_crispr_genome_editing",
        "optional_parameters": [
            {
                "default": None,
                "description": "The homology-directed repair template sequence, if used",
                "name": "repair_template",
                "type": "str",
            }
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "The original DNA sequence before CRISPR-Cas9 editing",
                "name": "original_sequence",
                "type": "str",
            },
            {
                "default": None,
                "description": "The DNA sequence after CRISPR-Cas9 editing",
                "name": "edited_sequence",
                "type": "str",
            },
            {
                "default": None,
                "description": "The CRISPR guide RNA (crRNA) sequence used for targeting",
                "name": "guide_rna",
                "type": "str",
            },
        ],
    },
    {
        "description": "Simulate DNA sequences with specified demographic and coalescent histories using msprime.",
        "name": "simulate_demographic_history",
        "optional_parameters": [
            {
                "default": 10,
                "description": "Number of sample sequences to simulate",
                "name": "num_samples",
                "type": "int",
            },
            {
                "default": 100000,
                "description": "Length of the simulated sequence in base pairs",
                "name": "sequence_length",
                "type": "int",
            },
            {
                "default": 1e-08,
                "description": "Per-base recombination rate",
                "name": "recombination_rate",
                "type": "float",
            },
            {
                "default": 1e-08,
                "description": "Per-base mutation rate",
                "name": "mutation_rate",
                "type": "float",
            },
            {
                "default": "constant",
                "description": "Type of demographic model to "
                "simulate (constant, bottleneck, "
                "expansion, contraction, sawtooth)",
                "name": "demographic_model",
                "type": "str",
            },
            {
                "default": None,
                "description": "Parameters specific to the chosen demographic model",
                "name": "demographic_params",
                "type": "dict",
            },
            {
                "default": "kingman",
                "description": "Type of coalescent model to use (kingman, beta)",
                "name": "coalescent_model",
                "type": "str",
            },
            {
                "default": None,
                "description": "Parameter for beta-coalescent model",
                "name": "beta_coalescent_param",
                "type": "float",
            },
            {
                "default": None,
                "description": "Seed for random number generator",
                "name": "random_seed",
                "type": "int",
            },
            {
                "default": "simulated_sequences.vcf",
                "description": "Filename to save the simulated sequences in VCF format",
                "name": "output_file",
                "type": "str",
            },
        ],
        "required_parameters": [],
    },
    {
        "description": "Identifies binding sites for a specific transcription factor in a genomic sequence.",
        "name": "identify_transcription_factor_binding_sites",
        "optional_parameters": [
            {
                "default": 0.8,
                "description": "Minimum score threshold for reporting binding sites (0.0-1.0)",
                "name": "threshold",
                "type": "float",
            },
            {
                "default": None,
                "description": "Path to save the results",
                "name": "output_file",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "The genomic DNA sequence to analyze",
                "name": "sequence",
                "type": "str",
            },
            {
                "default": None,
                "description": "Name of the transcription factor to search for (e.g., 'Hsf1', 'GATA1')",
                "name": "tf_name",
                "type": "str",
            },
        ],
    },
    {
        "description": "Fit a linear mixed model for genomic prediction using genotype and phenotype data.",
        "name": "fit_genomic_prediction_model",
        "optional_parameters": [
            {
                "default": None,
                "description": "Matrix of fixed effects (e.g., "
                "environment, management), with "
                "individuals in rows and effects in "
                "columns.",
                "name": "fixed_effects",
                "type": "numpy.ndarray",
            },
            {
                "default": "additive",
                "description": 'Type of genetic model to fit: "additive" or "additive_dominance".',
                "name": "model_type",
                "type": "str",
            },
            {
                "default": "genomic_prediction_results.csv",
                "description": "File name to save the results.",
                "name": "output_file",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Matrix of genotype data, with "
                "individuals in rows and markers in "
                "columns. Values are typically coded "
                "as 0, 1, 2 for additive models or "
                "with specific encoding for "
                "dominance effects.",
                "name": "genotypes",
                "type": "numpy.ndarray",
            },
            {
                "default": None,
                "description": "Vector or matrix of phenotype data, with individuals in rows and traits in columns.",
                "name": "phenotypes",
                "type": "numpy.ndarray",
            },
        ],
    },
    {
        "description": "Performs PCR amplification of a target transgene and "
        "visualizes results using agarose gel electrophoresis.",
        "name": "perform_pcr_and_gel_electrophoresis",
        "optional_parameters": [
            {
                "default": None,
                "description": "Forward primer sequence. If not provided, will be designed based on target_region",
                "name": "forward_primer",
                "type": "str",
            },
            {
                "default": None,
                "description": "Reverse primer sequence. If not provided, will be designed based on target_region",
                "name": "reverse_primer",
                "type": "str",
            },
            {
                "default": None,
                "description": "Tuple of (start, end) positions for the target region in the genomic DNA",
                "name": "target_region",
                "type": "tuple",
            },
            {
                "default": 58,
                "description": "Annealing temperature for PCR in °C",
                "name": "annealing_temp",
                "type": "float",
            },
            {
                "default": 30,
                "description": "Extension time in seconds",
                "name": "extension_time",
                "type": "int",
            },
            {
                "default": 35,
                "description": "Number of PCR cycles",
                "name": "cycles",
                "type": "int",
            },
            {
                "default": 2.0,
                "description": "Percentage of agarose gel",
                "name": "gel_percentage",
                "type": "float",
            },
            {
                "default": "pcr_result",
                "description": "Prefix for output files",
                "name": "output_prefix",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to file containing genomic DNA sequence in FASTA format or the sequence itself",
                "name": "genomic_dna",
                "type": "str",
            }
        ],
    },
    {
        "description": "Perform phylogenetic analysis on a set of protein sequences. "
        "This function aligns sequences, constructs a phylogenetic "
        "tree, and visualizes evolutionary relationships.",
        "name": "analyze_protein_phylogeny",
        "optional_parameters": [
            {
                "default": "./",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            },
            {
                "default": "clustalw",
                "description": 'Method for sequence alignment: "clustalw", "muscle", or "pre-aligned"',
                "name": "alignment_method",
                "type": "str",
            },
            {
                "default": "fasttree",
                "description": 'Method for tree construction: "iqtree" or fallback to neighbor-joining',
                "name": "tree_method",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to a FASTA file containing "
                "protein sequences or a string with "
                "FASTA-formatted sequences",
                "name": "fasta_sequences",
                "type": "str",
            }
        ],
    },
]



import os
from typing import Literal, Optional

import openai
from langchain_anthropic import ChatAnthropic

# from langchain_aws import ChatBedrock
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_google_genai import ChatGoogleGenerativeAI
from langchain_ollama import ChatOllama
from langchain_openai import AzureChatOpenAI, ChatOpenAI

SourceType = Literal["OpenAI", "AzureOpenAI", "Anthropic", "Ollama", "Gemini", "Bedrock", "Groq", "Custom"]
ALLOWED_SOURCES: set[str] = set(SourceType.__args__)


def get_llm(
    model: str = "claude-3-5-sonnet-20241022",
    temperature: float = 0.7,
    stop_sequences: list[str] | None = None,
    source: SourceType | None = None,
    base_url: str | None = None,
    api_key: str = "EMPTY",
) -> BaseChatModel:
    """
    Get a language model instance based on the specified model name and source.
    This function supports models from OpenAI, Azure OpenAI, Anthropic, Ollama, Gemini, Bedrock, and custom model serving.
    Args:
        model (str): The model name to use
        temperature (float): Temperature setting for generation
        stop_sequences (list): Sequences that will stop generation
        source (str): Source provider: "OpenAI", "AzureOpenAI", "Anthropic", "Ollama", "Gemini", "Bedrock", or "Custom"
                      If None, will attempt to auto-detect from model name
        base_url (str): The base URL for custom model serving (e.g., "http://localhost:8000/v1"), default is None
        api_key (str): The API key for the custom llm
    """
    # Auto-detect source from model name if not specified
    if source is None:
        env_source = os.getenv("LLM_SOURCE")
        if env_source in ALLOWED_SOURCES:
            source = env_source
        else:
            if model[:7] == "claude-":
                source = "Anthropic"
            elif model[:4] == "gpt-":
                source = "OpenAI"
            elif model.startswith("azure-"):
                source = "AzureOpenAI"
            elif model[:7] == "gemini-":
                source = "Gemini"
            elif "groq" in model.lower():
                source = "Groq"
            elif base_url is not None:
                source = "Custom"
            elif "/" in model or any(
                name in model.lower()
                for name in ["llama", "mistral", "qwen", "gemma", "phi", "dolphin", "orca", "vicuna", "deepseek"]
            ):
                source = "Ollama"
            elif model.startswith(
                ("anthropic.claude-", "amazon.titan-", "meta.llama-", "mistral.", "cohere.", "ai21.", "us.")
            ):
                source = "Bedrock"
            else:
                raise ValueError("Unable to determine model source. Please specify 'source' parameter.")

    # Create appropriate model based on source
    if source == "OpenAI":
        return ChatOpenAI(model=model, temperature=temperature, stop_sequences=stop_sequences)
    elif source == "AzureOpenAI":
        API_VERSION = "2024-12-01-preview"
        model = model.replace("azure-", "")
        return AzureChatOpenAI(
            openai_api_key=os.getenv("OPENAI_API_KEY"),
            azure_endpoint=os.getenv("OPENAI_ENDPOINT"),
            azure_deployment=model,
            openai_api_version=API_VERSION,
            temperature=temperature,
        )
    elif source == "Anthropic":
        return ChatAnthropic(
            model=model,
            temperature=temperature,
            max_tokens=8192,
            stop_sequences=stop_sequences,
        )
    elif source == "Gemini":
        # If you want to use ChatGoogleGenerativeAI, you need to pass the stop sequences upon invoking the model.
        # return ChatGoogleGenerativeAI(
        #     model=model,
        #     temperature=temperature,
        #     google_api_key=api_key,
        # )
        return ChatOpenAI(
            model=model,
            temperature=temperature,
            api_key=os.getenv("GEMINI_API_KEY"),
            base_url="https://generativelanguage.googleapis.com/v1beta/openai/",
            stop_sequences=stop_sequences,
        )
    elif source == "Groq":
        return ChatOpenAI(
            model=model,
            temperature=temperature,
            api_key=os.getenv("GROQ_API_KEY"),
            base_url="https://api.groq.com/openai/v1",
            stop_sequences=stop_sequences,
        )
    elif source == "Ollama":
        return ChatOllama(
            model=model,
            temperature=temperature,
        )

    # elif source == "Bedrock":
    #     return ChatBedrock(
    #         model=model,
    #         temperature=temperature,
    #         stop_sequences=stop_sequences,
    #         region_name=os.getenv("AWS_REGION", "us-east-1"),
    #     )
    elif source == "Custom":
        # Custom LLM serving such as SGLang. Must expose an openai compatible API.
        assert base_url is not None, "base_url must be provided for customly served LLMs"
        llm = ChatOpenAI(
            model=model,
            temperature=temperature,
            max_tokens=8192,
            stop_sequences=stop_sequences,
            base_url=base_url,
            api_key=api_key,
        )
        return llm
    else:
        raise ValueError(
            f"Invalid source: {source}. Valid options are 'OpenAI', 'AzureOpenAI', 'Anthropic', 'Gemini', 'Bedrock', or 'Ollama'"
        )



description = [
    {
        "description": "Analyze aortic diameter and geometry from cardiovascular "
        "imaging data to measure aortic root diameter, ascending "
        "aorta diameter, and calculate geometric parameters such as "
        "tortuosity and dilation indices.",
        "name": "analyze_aortic_diameter_and_geometry",
        "optional_parameters": [
            {
                "default": "./output",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            }
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the cardiovascular imaging data (DICOM, JPG, PNG)",
                "name": "image_path",
                "type": "str",
            }
        ],
    },
    {
        "description": "Analyze luminescence-based ATP assay data to determine intracellular ATP concentration.",
        "name": "analyze_atp_luminescence_assay",
        "optional_parameters": [
            {
                "default": "cell_count",
                "description": "Method used to normalize ATP values, either cell_count or protein_content",
                "name": "normalization_method",
                "type": "str",
            },
            {
                "default": None,
                "description": "Path to CSV file with normalization "
                "data or dictionary with sample IDs "
                "as keys and normalization values",
                "name": "normalization_data",
                "type": "str or dict",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to CSV file containing "
                "luminescence readings from samples "
                "with columns for Sample_ID and "
                "Luminescence_Value",
                "name": "data_file",
                "type": "str",
            },
            {
                "default": None,
                "description": "Path to CSV file containing "
                "standard curve data with columns "
                "for ATP_Concentration (in nM) and "
                "Luminescence_Value",
                "name": "standard_curve_file",
                "type": "str",
            },
        ],
    },
    {
        "description": "Analyze histological images of thrombus samples stained with "
        "H&E to identify and quantify different thrombus components "
        "(fresh, cellular lysis, endothelialization, fibroblastic "
        "reaction).",
        "name": "analyze_thrombus_histology",
        "optional_parameters": [
            {
                "default": "./output",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            }
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the histological image of thrombus sample stained with H&E",
                "name": "image_path",
                "type": "str",
            }
        ],
    },
    {
        "description": "Analyzes intracellular calcium concentration using Rhod-2 "
        "fluorescent indicator from microscopy images.",
        "name": "analyze_intracellular_calcium_with_rhod2",
        "optional_parameters": [
            {
                "default": "./output",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            }
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the background image (no cells, just media)",
                "name": "background_image_path",
                "type": "str",
            },
            {
                "default": None,
                "description": "Path to the control image (cells without calcium stimulus)",
                "name": "control_image_path",
                "type": "str",
            },
            {
                "default": None,
                "description": "Path to the sample image (cells with calcium stimulus)",
                "name": "sample_image_path",
                "type": "str",
            },
        ],
    },
    {
        "description": "Quantify the volume/density of immunofluorescence-labeled corneal nerve fibers.",
        "name": "quantify_corneal_nerve_fibers",
        "optional_parameters": [
            {
                "default": "./output",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            },
            {
                "default": "otsu",
                "description": "Method for thresholding ('otsu', 'adaptive', 'manual')",
                "name": "threshold_method",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the immunofluorescence microscopy image file",
                "name": "image_path",
                "type": "str",
            },
            {
                "default": None,
                "description": "Type of nerve fiber marker (e.g., 'βIII-tubulin', 'SP', 'L1CAM')",
                "name": "marker_type",
                "type": "str",
            },
        ],
    },
    {
        "description": "Segment cells and quantify protein expression levels from multichannel tissue images.",
        "name": "segment_and_quantify_cells_in_multiplexed_images",
        "optional_parameters": [
            {
                "default": 0,
                "description": "Index of the nuclear marker channel (typically DAPI)",
                "name": "nuclear_channel_index",
                "type": "int",
            },
            {
                "default": "./output",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the multichannel image file (tiff stack or similar format)",
                "name": "image_path",
                "type": "str",
            },
            {
                "default": None,
                "description": "List of marker names corresponding to each channel in the image",
                "name": "markers_list",
                "type": "List[str]",
            },
        ],
    },
    {
        "description": "Analyze bone microarchitecture parameters from 3D micro-CT "
        "images to calculate bone mineral density, bone volume, "
        "trabecular number, thickness, and separation.",
        "name": "analyze_bone_microct_morphometry",
        "optional_parameters": [
            {
                "default": "./results",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            },
            {
                "default": None,
                "description": "Threshold value for bone segmentation. If None, Otsu's method will be used",
                "name": "threshold_value",
                "type": "float",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the micro-CT scan data file (TIFF stack or similar 3D format)",
                "name": "input_file_path",
                "type": "str",
            }
        ],
    },
]



description = [
    {
        "description": "Analyze DNA Damage Response (DDR) network alterations and dependencies in cancer samples.",
        "name": "analyze_ddr_network_in_cancer",
        "optional_parameters": [
            {
                "default": "./results",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            }
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to gene expression data file (CSV format with genes as rows, samples as columns)",
                "name": "expression_data_path",
                "type": "str",
            },
            {
                "default": None,
                "description": "Path to mutation data file (CSV "
                "format with genes as rows, samples "
                "as columns, values indicating "
                "mutation status)",
                "name": "mutation_data_path",
                "type": "str",
            },
        ],
    },
    {
        "description": "Analyze flow cytometry data to quantify senescent and apoptotic cell populations.",
        "name": "analyze_cell_senescence_and_apoptosis",
        "optional_parameters": [],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the FCS file containing "
                "flow cytometry data with "
                "measurements for "
                "senescence-associated "
                "β-galactosidase (SA-β-Gal) and "
                "Annexin V/7-AAD staining",
                "name": "fcs_file_path",
                "type": "str",
            }
        ],
    },
    {
        "description": "Detects and annotates somatic mutations in tumor samples "
        "compared to matched normal samples using GATK Mutect2 for "
        "variant calling, GATK FilterMutectCalls for filtering, and "
        "SnpEff for functional annotation.",
        "name": "detect_and_annotate_somatic_mutations",
        "optional_parameters": [
            {
                "default": "GRCh38.105",
                "description": "SnpEff database to use for annotation",
                "name": "snpeff_database",
                "type": "str",
            }
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the tumor sample BAM file",
                "name": "tumor_bam",
                "type": "str",
            },
            {
                "default": None,
                "description": "Path to the matched normal sample BAM file",
                "name": "normal_bam",
                "type": "str",
            },
            {
                "default": None,
                "description": "Path to the reference genome FASTA file",
                "name": "reference_genome",
                "type": "str",
            },
            {
                "default": None,
                "description": "Prefix for output files",
                "name": "output_prefix",
                "type": "str",
            },
        ],
    },
    {
        "description": "Detects and characterizes structural variations (SVs) in "
        "genomic sequencing data using LUMPY for SV detection "
        "followed by annotation with COSMIC and/or ClinVar databases.",
        "name": "detect_and_characterize_structural_variations",
        "optional_parameters": [
            {
                "default": None,
                "description": "Path to the COSMIC database for cancer annotation",
                "name": "cosmic_db_path",
                "type": "str",
            },
            {
                "default": None,
                "description": "Path to the ClinVar database for clinical annotation",
                "name": "clinvar_db_path",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the aligned sequencing data in BAM format",
                "name": "bam_file_path",
                "type": "str",
            },
            {
                "default": None,
                "description": "Path to the reference genome in FASTA format",
                "name": "reference_genome_path",
                "type": "str",
            },
            {
                "default": None,
                "description": "Directory where results will be saved",
                "name": "output_dir",
                "type": "str",
            },
        ],
    },
    {
        "description": "Performs Non-negative Matrix Factorization (NMF) on gene "
        "expression data to extract metagenes and their associated "
        "sample weights for tumor subtype identification.",
        "name": "perform_gene_expression_nmf_analysis",
        "optional_parameters": [
            {
                "default": 10,
                "description": "Number of metagenes (components) to extract.",
                "name": "n_components",
                "type": "int",
            },
            {
                "default": True,
                "description": "Whether to normalize the expression data before applying NMF.",
                "name": "normalize",
                "type": "bool",
            },
            {
                "default": "nmf_results",
                "description": "Directory to save the output files.",
                "name": "output_dir",
                "type": "str",
            },
            {
                "default": 42,
                "description": "Random seed for reproducibility.",
                "name": "random_state",
                "type": "int",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to a CSV or TSV file "
                "containing gene expression data "
                "with genes as rows and samples as "
                "columns. Values should be "
                "non-negative.",
                "name": "expression_data_path",
                "type": "str",
            }
        ],
    },
]



def quantify_cell_cycle_phases_from_microscopy(image_paths, output_dir="./results"):
    """Quantify the percentage of cells in each cell cycle phase using Calcofluor white stained microscopy images.

    This function processes microscopy images where cell walls/septa are stained with Calcofluor white,
    segments individual cells, extracts features, and classifies each cell into G1, S, or G2/M phase.

    Parameters
    ----------
    image_paths : list of str
        List of file paths to microscopy images of cells stained with Calcofluor white
    output_dir : str, optional
        Directory to save results (default: './results')

    Returns
    -------
    str
        Research log summarizing the analysis process and results

    """
    import os

    import numpy as np
    import pandas as pd
    from skimage import filters, io, measure, morphology

    # Create output directory if it doesn't exist
    os.makedirs(output_dir, exist_ok=True)

    # Initialize log
    log = "# Cell Cycle Phase Quantification Analysis Log\n\n"
    log += f"Processing {len(image_paths)} microscopy images with Calcofluor white staining.\n\n"

    # Initialize data structures to store results
    all_cells = []
    cell_features = []

    # Step 1: Process images and segment cells
    log += "## Image Processing and Cell Segmentation\n\n"

    for i, img_path in enumerate(image_paths):
        log += f"Processing image {i + 1}/{len(image_paths)}: {os.path.basename(img_path)}\n"

        # Load image
        img = io.imread(img_path)

        # Convert to grayscale if needed
        if len(img.shape) > 2:
            img = np.mean(img, axis=2).astype(np.uint8)

        # Preprocessing
        filtered_img = filters.gaussian(img, sigma=1.0)

        # Thresholding to identify cell walls
        threshold = filters.threshold_otsu(filtered_img)
        binary = filtered_img > threshold

        # Clean up binary image
        binary = morphology.remove_small_objects(binary, min_size=30)
        binary = morphology.binary_closing(binary)

        # Segment cells
        labeled_cells = measure.label(binary)
        cell_props = measure.regionprops(labeled_cells, intensity_image=img)

        log += f"  - Found {len(cell_props)} cells\n"
        all_cells.extend(cell_props)

        # Extract features for each cell
        for cell in cell_props:
            # Basic morphological features
            area = cell.area
            perimeter = cell.perimeter
            eccentricity = cell.eccentricity
            mean_intensity = cell.mean_intensity

            # Cell shape features
            circularity = (4 * np.pi * area) / (perimeter**2) if perimeter > 0 else 0

            # Check for septation (indicative of S or G2/M phase)
            # Calcofluor white stains the septum more intensely
            has_septum = False
            if area > 50:  # Only check larger cells
                intensity_profile = cell.intensity_image[cell.intensity_image > 0]
                if intensity_profile.size > 0:
                    intensity_std = np.std(intensity_profile)
                    intensity_max = np.max(intensity_profile)
                    # High standard deviation and max intensity suggests septum presence
                    has_septum = intensity_std > 0.2 * np.mean(intensity_profile) and intensity_max > 1.5 * np.mean(
                        intensity_profile
                    )

            # Store features
            cell_features.append(
                [
                    area,
                    perimeter,
                    eccentricity,
                    mean_intensity,
                    circularity,
                    has_septum,
                    i,  # include image index
                ]
            )

    log += f"\nTotal cells detected across all images: {len(all_cells)}\n\n"

    # Step 2: Feature engineering and classification
    log += "## Cell Cycle Phase Classification\n\n"

    # Convert to numpy array for easier processing
    X = np.array(cell_features)

    # Simple rule-based classification
    # This is a simplified approach - in practice, you would train a classifier on labeled data
    # G1: Smaller cells, no septum
    # S: Medium-sized cells, early septum formation
    # G2/M: Larger cells, clear septum, often elongated

    # Define classification rules based on features
    phases = []
    for i in range(X.shape[0]):
        area = X[i, 0]
        perimeter = X[i, 1]
        eccentricity = X[i, 2]
        has_septum = X[i, 5]

        if has_septum and area > np.median(X[:, 0]) * 1.2:
            # Larger cells with septum are likely in G2/M
            phases.append("G2/M")
        elif has_septum or (area > np.median(X[:, 0]) and eccentricity > 0.5):
            # Cells with septum or larger elongated cells are likely in S
            phases.append("S")
        else:
            # Smaller cells without septum are likely in G1
            phases.append("G1")

    # Calculate percentages
    unique_phases, counts = np.unique(phases, return_counts=True)
    percentages = (counts / len(phases)) * 100

    # Create results table
    results_df = pd.DataFrame({"Phase": unique_phases, "Count": counts, "Percentage": percentages})

    # Save results
    results_path = os.path.join(output_dir, "cell_cycle_phases.csv")
    results_df.to_csv(results_path, index=False)

    # Log results
    log += "### Cell Cycle Phase Distribution\n\n"
    for phase, count, percentage in zip(unique_phases, counts, percentages, strict=False):
        log += f"- {phase}: {count} cells ({percentage:.2f}%)\n"

    log += f"\nResults saved to: {results_path}\n"
    log += "\nNote: This analysis uses simplified morphological features for classification. "
    log += (
        "For more accurate results, a supervised machine learning approach with labeled training data is recommended.\n"
    )

    return log


def quantify_and_cluster_cell_motility(image_sequence_path, output_dir="./results", num_clusters=3):
    """Quantify cell motility features from time-lapse microscopy images and cluster cells based on motility patterns.

    Parameters
    ----------
    image_sequence_path : str
        Path to directory containing time-lapse microscopy images in sequential order
    output_dir : str
        Directory to save output files (default: "./results")
    num_clusters : int
        Number of motility pattern clusters to identify (default: 3)

    Returns
    -------
    str
        Research log summarizing the analysis process and results

    """
    import os

    import cv2
    import numpy as np
    import pandas as pd
    from sklearn.cluster import KMeans

    # Create output directory if it doesn't exist
    os.makedirs(output_dir, exist_ok=True)

    # Step 1: Load image sequence
    image_files = sorted(
        [f for f in os.listdir(image_sequence_path) if f.endswith((".tif", ".tiff", ".png", ".jpg", ".jpeg"))]
    )

    if len(image_files) < 2:
        return "Error: Insufficient images found. At least 2 time points are required."

    # Step 2: Initialize cell tracking
    first_image = cv2.imread(os.path.join(image_sequence_path, image_files[0]), cv2.IMREAD_GRAYSCALE)

    # Simple cell detection using thresholding and contour detection
    _, binary = cv2.threshold(first_image, 127, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
    contours, _ = cv2.findContours(binary, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    # Extract cell centroids from first frame
    cells = []
    for i, contour in enumerate(contours):
        if cv2.contourArea(contour) > 50:  # Filter small noise
            M = cv2.moments(contour)
            if M["m00"] != 0:
                cx = int(M["m10"] / M["m00"])
                cy = int(M["m01"] / M["m00"])
                cells.append({"id": i, "positions": [(cx, cy)], "frame_indices": [0]})

    # Step 3: Track cells across frames
    for frame_idx, img_file in enumerate(image_files[1:], 1):
        img = cv2.imread(os.path.join(image_sequence_path, img_file), cv2.IMREAD_GRAYSCALE)
        _, binary = cv2.threshold(img, 127, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
        contours, _ = cv2.findContours(binary, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

        # Get current frame centroids
        current_centroids = []
        for contour in contours:
            if cv2.contourArea(contour) > 50:
                M = cv2.moments(contour)
                if M["m00"] != 0:
                    cx = int(M["m10"] / M["m00"])
                    cy = int(M["m01"] / M["m00"])
                    current_centroids.append((cx, cy))

        # Match cells with nearest centroids in current frame
        for cell in cells:
            if len(cell["positions"]) == frame_idx:  # Only track cells found in previous frame
                prev_pos = cell["positions"][-1]

                if current_centroids:
                    # Calculate distances to all current centroids
                    distances = [
                        np.sqrt((prev_pos[0] - c[0]) ** 2 + (prev_pos[1] - c[1]) ** 2) for c in current_centroids
                    ]
                    min_idx = np.argmin(distances)

                    # Only match if distance is below threshold (prevents incorrect matching)
                    if distances[min_idx] < 50:  # Threshold distance in pixels
                        cell["positions"].append(current_centroids[min_idx])
                        cell["frame_indices"].append(frame_idx)
                        current_centroids.pop(min_idx)  # Remove matched centroid

    # Step 4: Calculate motility features for each cell
    cell_features = []

    for cell in cells:
        # Only analyze cells tracked for at least 3 frames
        if len(cell["positions"]) < 3:
            continue

        # Calculate displacements between consecutive positions
        displacements = []
        for i in range(1, len(cell["positions"])):
            prev_pos = cell["positions"][i - 1]
            curr_pos = cell["positions"][i]
            displacement = np.sqrt((curr_pos[0] - prev_pos[0]) ** 2 + (curr_pos[1] - prev_pos[1]) ** 2)
            displacements.append(displacement)

        # Calculate speed (pixels per frame)
        avg_speed = np.mean(displacements)

        # Calculate directionality (displacement from start to end / total path length)
        start_pos = cell["positions"][0]
        end_pos = cell["positions"][-1]
        net_displacement = np.sqrt((end_pos[0] - start_pos[0]) ** 2 + (end_pos[1] - start_pos[1]) ** 2)
        total_path_length = np.sum(displacements)
        directionality = net_displacement / total_path_length if total_path_length > 0 else 0

        # Calculate mean squared displacement
        msd = np.mean(
            [
                np.sqrt((cell["positions"][i][0] - start_pos[0]) ** 2 + (cell["positions"][i][1] - start_pos[1]) ** 2)
                for i in range(1, len(cell["positions"]))
            ]
        )

        # Calculate track duration
        duration = cell["frame_indices"][-1] - cell["frame_indices"][0]

        cell_features.append(
            {
                "cell_id": cell["id"],
                "avg_speed": avg_speed,
                "directionality": directionality,
                "msd": msd,
                "track_duration": duration,
                "track_length": len(cell["positions"]),
            }
        )

    # Step 5: Cluster cells based on motility features
    if len(cell_features) < num_clusters:
        return f"Error: Not enough cells ({len(cell_features)}) to form {num_clusters} clusters. Try reducing num_clusters."

    # Create feature matrix for clustering
    feature_df = pd.DataFrame(cell_features)
    feature_matrix = feature_df[["avg_speed", "directionality", "msd"]].values

    # Normalize features
    feature_matrix = (feature_matrix - np.mean(feature_matrix, axis=0)) / np.std(feature_matrix, axis=0)

    # Perform k-means clustering
    kmeans = KMeans(n_clusters=num_clusters, random_state=42)
    clusters = kmeans.fit_predict(feature_matrix)

    # Add cluster assignments to dataframe
    feature_df["cluster"] = clusters

    # Save results to CSV
    results_file = os.path.join(output_dir, "cell_motility_features.csv")
    feature_df.to_csv(results_file, index=False)

    # Calculate cluster statistics
    cluster_stats = feature_df.groupby("cluster").agg(
        {
            "avg_speed": ["mean", "std"],
            "directionality": ["mean", "std"],
            "msd": ["mean", "std"],
            "cell_id": "count",
        }
    )

    # Save cluster statistics
    stats_file = os.path.join(output_dir, "cluster_statistics.csv")
    cluster_stats.to_csv(stats_file)

    # Create research log
    log = f"""
Cell Motility Quantification and Clustering Analysis
===================================================

Analysis Steps:
1. Loaded {len(image_files)} time-lapse microscopy images
2. Detected and tracked {len(cells)} initial cells across time frames
3. Calculated motility features for {len(cell_features)} cells with sufficient tracking data
4. Performed k-means clustering to identify {num_clusters} distinct motility patterns

Results Summary:
- Detected {len(cells)} cells in the first frame
- Successfully tracked {len(cell_features)} cells across multiple frames
- Clustered cells into {num_clusters} motility pattern groups

Cluster Statistics:
{cluster_stats.to_string()}

Output Files:
- Cell motility features saved to: {results_file}
- Cluster statistics saved to: {stats_file}

Analysis Notes:
- Features used for clustering: average speed, directionality, and mean squared displacement
- Cells were required to be tracked for at least 3 frames to be included in the analysis
"""

    # Save research log
    log_file = os.path.join(output_dir, "research_log.txt")
    with open(log_file, "w") as f:
        f.write(log)

    return log


def perform_facs_cell_sorting(
    cell_suspension_data,
    fluorescence_parameter,
    threshold_min=None,
    threshold_max=None,
    output_file="sorted_cells.csv",
):
    """Performs Fluorescence-Activated Cell Sorting (FACS) to enrich cell populations based on fluorescence characteristics.

    Parameters
    ----------
    cell_suspension_data : str
        Path to the FCS file containing flow cytometry data
    fluorescence_parameter : str
        The fluorescence parameter to use for sorting (e.g., 'GFP', 'FITC', 'PE')
    threshold_min : float, optional
        Minimum threshold for the fluorescence parameter. Cells below this value will be excluded
    threshold_max : float, optional
        Maximum threshold for the fluorescence parameter. Cells above this value will be excluded
    output_file : str, optional
        Filename to save the sorted cell population data

    Returns
    -------
    str
        Research log detailing the FACS cell sorting process

    """
    import os

    import pandas as pd

    # Initialize research log
    log = "# FACS-based Cell Sorting and Enrichment Research Log\n\n"

    try:
        # Load cell suspension data
        log += "## Loading Cell Suspension Data\n"
        if isinstance(cell_suspension_data, str):
            # If data is provided as a file path
            if cell_suspension_data.lower().endswith(".fcs"):
                try:
                    import flowkit as fk

                    fcs_data = fk.Sample(cell_suspension_data)
                    cell_df = pd.DataFrame(fcs_data.get_dataframe())
                    log += f"Successfully loaded FCS file containing {len(cell_df)} cells\n"
                except ImportError:
                    # Fallback if flowkit is not available
                    log += "FlowKit not available, attempting to load as CSV\n"
                    cell_df = pd.read_csv(cell_suspension_data)
            else:
                # Assume CSV or other tabular format
                cell_df = pd.read_csv(cell_suspension_data)
                log += f"Loaded data file containing {len(cell_df)} cells\n"
        else:
            # Assume data is already a DataFrame
            cell_df = cell_suspension_data.copy()
            log += f"Using provided DataFrame containing {len(cell_df)} cells\n"

        # Validate fluorescence parameter exists in data
        if fluorescence_parameter not in cell_df.columns:
            raise ValueError(f"Fluorescence parameter '{fluorescence_parameter}' not found in data")

        # Apply gating strategy based on fluorescence thresholds
        log += f"\n## Applying Gating Strategy for {fluorescence_parameter}\n"
        original_count = len(cell_df)

        # Filter cells based on min/max thresholds
        if threshold_min is not None:
            cell_df = cell_df[cell_df[fluorescence_parameter] >= threshold_min]
            log += f"Applied minimum threshold of {threshold_min}: {len(cell_df)} cells remaining\n"

        if threshold_max is not None:
            cell_df = cell_df[cell_df[fluorescence_parameter] <= threshold_max]
            log += f"Applied maximum threshold of {threshold_max}: {len(cell_df)} cells remaining\n"

        # Calculate enrichment statistics
        enrichment_percent = (len(cell_df) / original_count) * 100 if original_count > 0 else 0
        log += "\n## Enrichment Results\n"
        log += f"Original population: {original_count} cells\n"
        log += f"Sorted population: {len(cell_df)} cells\n"
        log += f"Enrichment percentage: {enrichment_percent:.2f}%\n"

        # Calculate statistics of sorted population
        mean_fluorescence = cell_df[fluorescence_parameter].mean()
        median_fluorescence = cell_df[fluorescence_parameter].median()

        log += f"Mean {fluorescence_parameter}: {mean_fluorescence:.2f}\n"
        log += f"Median {fluorescence_parameter}: {median_fluorescence:.2f}\n"

        # Save sorted population to file
        cell_df.to_csv(output_file, index=False)
        log += "\n## Output\n"
        log += f"Sorted cell population saved to: {os.path.abspath(output_file)}\n"

        return log

    except Exception as e:
        log += "\n## Error\n"
        log += f"An error occurred during cell sorting: {str(e)}\n"
        return log


def analyze_flow_cytometry_immunophenotyping(
    fcs_file_path, gating_strategy, compensation_matrix=None, output_dir="./results"
):
    """Analyze flow cytometry data to identify and quantify specific cell populations based on surface markers.

    Parameters
    ----------
    fcs_file_path : str
        Path to the FCS file containing flow cytometry data
    gating_strategy : dict
        Dictionary defining the gating strategy. Each key is a population name, and each value is a list of tuples
        (marker, operator, threshold). For example: {'HSCs': [('Lin', '<', 100), ('Sca1', '>', 1000), ...]}
    compensation_matrix : numpy.ndarray, optional
        Spillover/compensation matrix to correct for fluorescence overlap
    output_dir : str, optional
        Directory to save the results

    Returns
    -------
    str
        Research log summarizing the analysis steps and results

    """
    import os

    import pandas as pd
    from FlowCytometryTools import FCMeasurement

    # Create output directory if it doesn't exist
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # Load the FCS file
    sample = FCMeasurement(ID="Sample", datafile=fcs_file_path)

    # Apply compensation if provided
    if compensation_matrix is not None:
        sample = sample.compensate(compensation_matrix)

    # Initialize log
    log = "Flow Cytometry Analysis Log\n"
    log += "==========================\n"
    log += f"File analyzed: {os.path.basename(fcs_file_path)}\n"
    log += f"Total events: {len(sample)}\n\n"

    # Create a dictionary to store population counts
    populations = {}

    # Apply gating strategy to identify cell populations
    for pop_name, gates in gating_strategy.items():
        # Start with all events
        current_population = sample.copy()

        log += f"Identifying {pop_name} population:\n"

        # Apply each gate sequentially
        for marker, operator, threshold in gates:
            initial_count = len(current_population)

            if operator == ">":
                current_population = current_population.gate(f"{marker} > {threshold}")
            elif operator == "<":
                current_population = current_population.gate(f"{marker} < {threshold}")
            elif operator == "between":
                # For 'between', threshold should be a tuple (lower, upper)
                lower, upper = threshold
                current_population = current_population.gate(f"{marker} > {lower} and {marker} < {upper}")

            final_count = len(current_population)
            log += f"  Gate {marker} {operator} {threshold}: {initial_count} → {final_count} events\n"

        # Store the final population
        populations[pop_name] = current_population

        # Calculate percentage of original sample
        percentage = (len(current_population) / len(sample)) * 100
        log += f"  Final {pop_name} count: {len(current_population)} events ({percentage:.2f}% of total)\n\n"

    # Create a summary dataframe
    summary_data = {"Population": [], "Count": [], "Percentage": []}

    for pop_name, population in populations.items():
        summary_data["Population"].append(pop_name)
        summary_data["Count"].append(len(population))
        summary_data["Percentage"].append((len(population) / len(sample)) * 100)

    summary_df = pd.DataFrame(summary_data)

    # Save the summary to a CSV file
    summary_file = os.path.join(output_dir, "population_summary.csv")
    summary_df.to_csv(summary_file, index=False)

    log += "Summary of identified populations:\n"
    for _, row in summary_df.iterrows():
        log += f"  {row['Population']}: {row['Count']} events ({row['Percentage']:.2f}%)\n"

    log += f"\nDetailed results saved to: {summary_file}\n"

    return log


def analyze_mitochondrial_morphology_and_potential(morphology_image_path, potential_image_path, output_dir="./output"):
    """Quantifies metrics of mitochondrial morphology and membrane potential from fluorescence microscopy images.

    Parameters
    ----------
    morphology_image_path : str
        Path to the fluorescence microscopy image showing mitochondrial morphology (e.g., MTS-GFP)
    potential_image_path : str
        Path to the fluorescence microscopy image showing mitochondrial membrane potential (e.g., TMRE staining)
    output_dir : str, optional
        Directory to save output files, default is "./output"

    Returns
    -------
    str
        A research log summarizing the analysis steps and results

    """
    import datetime
    import os

    import cv2
    import numpy as np
    from scipy import ndimage
    from skimage import filters, io, morphology, util

    # Create output directory if it doesn't exist
    os.makedirs(output_dir, exist_ok=True)

    log = []
    log.append(f"Mitochondrial Analysis - {datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}")
    log.append("=" * 50)

    # Load images
    try:
        morph_img = io.imread(morphology_image_path)
        pot_img = io.imread(potential_image_path)

        log.append(f"Successfully loaded morphology image: {morphology_image_path}")
        log.append(f"Successfully loaded potential image: {potential_image_path}")
        log.append(f"Morphology image shape: {morph_img.shape}")
        log.append(f"Potential image shape: {pot_img.shape}")
    except Exception as e:
        return f"Error loading images: {str(e)}"

    # Convert to grayscale if needed
    if len(morph_img.shape) > 2:
        morph_img = cv2.cvtColor(morph_img, cv2.COLOR_RGB2GRAY)
        log.append("Converted morphology image to grayscale")
    if len(pot_img.shape) > 2:
        pot_img = cv2.cvtColor(pot_img, cv2.COLOR_RGB2GRAY)
        log.append("Converted potential image to grayscale")

    # Normalize images
    morph_img = util.img_as_float(morph_img)
    pot_img = util.img_as_float(pot_img)

    log.append("\nMorphology Analysis:")
    log.append("-" * 30)

    # Denoise morphology image
    morph_img_denoised = filters.gaussian(morph_img, sigma=1.0)

    # Threshold to segment mitochondria
    threshold_value = filters.threshold_otsu(morph_img_denoised)
    binary_img = morph_img_denoised > threshold_value

    # Remove small objects
    binary_img = morphology.remove_small_objects(binary_img, min_size=20)

    # Save binary image
    binary_path = os.path.join(output_dir, "binary_mitochondria.png")
    io.imsave(binary_path, util.img_as_ubyte(binary_img))
    log.append(f"Binary segmentation saved to: {binary_path}")

    # Skeletonize for network analysis
    skeleton = morphology.skeletonize(binary_img)
    skeleton_path = os.path.join(output_dir, "skeleton.png")
    io.imsave(skeleton_path, util.img_as_ubyte(skeleton))
    log.append(f"Skeleton image saved to: {skeleton_path}")

    # Calculate morphology metrics
    # 1. Count branches and junctions
    # Label the skeleton
    labeled_skeleton, num_branches = ndimage.label(skeleton)

    # Find branch points (pixels with more than 2 neighbors)
    kernel = np.ones((3, 3), dtype=np.uint8)
    kernel[1, 1] = 0  # Don't count the center pixel
    neighbor_count = ndimage.convolve(skeleton.astype(np.uint8), kernel)
    junction_points = (neighbor_count > 2) & skeleton
    num_junctions = np.sum(junction_points)

    # 2. Calculate fragmentation metrics
    labeled_objects, num_objects = ndimage.label(binary_img)
    object_sizes = ndimage.sum(binary_img, labeled_objects, range(1, num_objects + 1))
    mean_size = np.mean(object_sizes) if len(object_sizes) > 0 else 0

    # Calculate network connectivity (ratio of junctions to branches)
    connectivity = num_junctions / num_branches if num_branches > 0 else 0

    # Calculate fragmentation index (inverse of mean object size, normalized)
    fragmentation = 1 / (mean_size / np.max(object_sizes)) if mean_size > 0 and np.max(object_sizes) > 0 else 0

    log.append(f"Number of mitochondrial fragments: {num_objects}")
    log.append(f"Number of branches: {num_branches}")
    log.append(f"Number of junction points: {num_junctions}")
    log.append(f"Network connectivity index: {connectivity:.4f}")
    log.append(f"Fragmentation index: {fragmentation:.4f}")
    log.append(f"Mean fragment size: {mean_size:.2f} pixels")

    # Membrane potential analysis
    log.append("\nMembrane Potential Analysis:")
    log.append("-" * 30)

    # Create mask from morphology image to analyze only mitochondrial regions
    mito_mask = binary_img

    # Calculate potential metrics within mitochondrial regions
    pot_intensity_raw = np.mean(pot_img)
    pot_intensity_in_mito = np.mean(pot_img[mito_mask]) if np.sum(mito_mask) > 0 else 0

    # Calculate potential heterogeneity (standard deviation of intensity)
    pot_heterogeneity = np.std(pot_img[mito_mask]) if np.sum(mito_mask) > 0 else 0

    # Calculate potential distribution
    if np.sum(mito_mask) > 0:
        percentiles = [10, 25, 50, 75, 90]
        pot_percentiles = np.percentile(pot_img[mito_mask], percentiles)
        percentile_str = ", ".join([f"{p}th: {v:.4f}" for p, v in zip(percentiles, pot_percentiles, strict=False)])
    else:
        percentile_str = "No mitochondrial regions detected"

    log.append(f"Overall mean TMRE intensity: {pot_intensity_raw:.4f}")
    log.append(f"Mean TMRE intensity in mitochondria: {pot_intensity_in_mito:.4f}")
    log.append(f"TMRE intensity heterogeneity (std): {pot_heterogeneity:.4f}")
    log.append(f"TMRE intensity percentiles: {percentile_str}")

    # Create a combined results image
    # Overlay potential intensity on morphology
    overlay = np.zeros((morph_img.shape[0], morph_img.shape[1], 3), dtype=np.float32)
    overlay[..., 0] = morph_img  # Red channel - morphology
    overlay[..., 2] = pot_img  # Blue channel - potential
    overlay = np.clip(overlay, 0, 1)

    overlay_path = os.path.join(output_dir, "morphology_potential_overlay.png")
    io.imsave(overlay_path, util.img_as_ubyte(overlay))
    log.append(f"Overlay image saved to: {overlay_path}")

    # Save quantitative results to CSV
    import csv

    results_path = os.path.join(output_dir, "mitochondrial_analysis_results.csv")
    with open(results_path, "w", newline="") as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(["Metric", "Value"])
        writer.writerow(["Number of fragments", num_objects])
        writer.writerow(["Number of branches", num_branches])
        writer.writerow(["Number of junctions", num_junctions])
        writer.writerow(["Network connectivity", connectivity])
        writer.writerow(["Fragmentation index", fragmentation])
        writer.writerow(["Mean fragment size", mean_size])
        writer.writerow(["Mean TMRE intensity", pot_intensity_in_mito])
        writer.writerow(["TMRE heterogeneity", pot_heterogeneity])

    log.append(f"Quantitative results saved to: {results_path}")
    log.append("\nAnalysis complete.")

    return "\n".join(log)



description = [
    {
        "description": "Generate a 3D model of facial anatomy from MRI scans of the head and neck.",
        "name": "reconstruct_3d_face_from_mri",
        "optional_parameters": [
            {
                "default": "./output",
                "description": "Directory where output files will be saved",
                "name": "output_dir",
                "type": "str",
            },
            {
                "default": "subject",
                "description": "Identifier for the subject, used in output filenames",
                "name": "subject_id",
                "type": "str",
            },
            {
                "default": 300,
                "description": "Threshold value for initial segmentation of facial tissues",
                "name": "threshold_value",
                "type": "int",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the MRI scan file (NIfTI format: .nii or .nii.gz)",
                "name": "mri_file_path",
                "type": "str",
            }
        ],
    },
    {
        "description": "Extracts P1 amplitude and latency from Auditory Brainstem Response (ABR) waveform data.",
        "name": "analyze_abr_waveform_p1_metrics",
        "optional_parameters": [],
        "required_parameters": [
            {
                "default": None,
                "description": "Time points of the ABR recording in milliseconds",
                "name": "time_ms",
                "type": "array-like",
            },
            {
                "default": None,
                "description": "Amplitude values of the ABR recording in microvolts",
                "name": "amplitude_uv",
                "type": "array-like",
            },
        ],
    },
    {
        "description": "Analyze ciliary beat frequency from high-speed video microscopy data using FFT analysis.",
        "name": "analyze_ciliary_beat_frequency",
        "optional_parameters": [
            {
                "default": 5,
                "description": "Number of regions of interest to analyze",
                "name": "roi_count",
                "type": "int",
            },
            {
                "default": 0,
                "description": "Minimum frequency to consider in Hz",
                "name": "min_freq",
                "type": "float",
            },
            {
                "default": 30,
                "description": "Maximum frequency to consider in Hz",
                "name": "max_freq",
                "type": "float",
            },
            {
                "default": "./",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the high-speed video microscopy file of ciliary beating",
                "name": "video_path",
                "type": "str",
            }
        ],
    },
    {
        "description": "Analyze colocalization between two fluorescently labeled proteins in microscopy images.",
        "name": "analyze_protein_colocalization",
        "optional_parameters": [
            {
                "default": "./output",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            },
            {
                "default": "otsu",
                "description": "Method for thresholding images ('otsu', 'li', or 'yen')",
                "name": "threshold_method",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the first channel image file (fluorescent protein 1)",
                "name": "channel1_path",
                "type": "str",
            },
            {
                "default": None,
                "description": "Path to the second channel image file (fluorescent protein 2)",
                "name": "channel2_path",
                "type": "str",
            },
        ],
    },
    {
        "description": "Performs cosinor analysis on physiological time series data to characterize circadian rhythms.",
        "name": "perform_cosinor_analysis",
        "optional_parameters": [
            {
                "default": 24.0,
                "description": "Period of the rhythm in hours, default is 24 hours for circadian rhythms",
                "name": "period",
                "type": "float",
            }
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Time points of the measurements in hours",
                "name": "time_data",
                "type": "array-like",
            },
            {
                "default": None,
                "description": "Physiological measurements corresponding to each time point",
                "name": "physiological_data",
                "type": "array-like",
            },
        ],
    },
    {
        "description": "Calculate Apparent Diffusion Coefficient (ADC) map from "
        "diffusion-weighted MRI data using monoexponential diffusion "
        "model.",
        "name": "calculate_brain_adc_map",
        "optional_parameters": [
            {
                "default": "adc_map.nii.gz",
                "description": "Path where the output ADC map will be saved",
                "name": "output_path",
                "type": "str",
            },
            {
                "default": None,
                "description": "Path to a binary mask file to limit ADC calculation to brain regions",
                "name": "mask_file_path",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the 4D NIfTI file containing diffusion-weighted MRI data",
                "name": "dwi_file_path",
                "type": "str",
            },
            {
                "default": None,
                "description": "List of b-values corresponding to each volume in the 4D DWI data",
                "name": "b_values",
                "type": "List[float]",
            },
        ],
    },
    {
        "description": "Analyze calcium dynamics in endo-lysosomal compartments using ELGA/ELGA1 probe data.",
        "name": "analyze_endolysosomal_calcium_dynamics",
        "optional_parameters": [
            {
                "default": None,
                "description": "Time point (in seconds) when treatment/stimulus was applied",
                "name": "treatment_time",
                "type": "float",
            },
            {
                "default": "",
                "description": "Type of cells used in the experiment",
                "name": "cell_type",
                "type": "str",
            },
            {
                "default": "",
                "description": "Name of the treatment or stimulus applied",
                "name": "treatment_name",
                "type": "str",
            },
            {
                "default": "calcium_analysis_results.txt",
                "description": "Name of the file to save detailed analysis results",
                "name": "output_file",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Time points of the measurements in seconds",
                "name": "time_points",
                "type": "numpy.ndarray or list",
            },
            {
                "default": None,
                "description": "Luminescence intensity values from ELGA/ELGA1 probes corresponding to Ca2+ levels",
                "name": "luminescence_values",
                "type": "numpy.ndarray or list",
            },
        ],
    },
    {
        "description": "Analyzes fatty acid composition in tissue samples using gas chromatography data.",
        "name": "analyze_fatty_acid_composition_by_gc",
        "optional_parameters": [
            {
                "default": "./results",
                "description": "Directory where result files will be saved",
                "name": "output_directory",
                "type": "str",
            }
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the CSV file containing gas "
                "chromatography data with columns "
                "'retention_time' and 'peak_area'",
                "name": "gc_data_file",
                "type": "str",
            },
            {
                "default": None,
                "description": "Type of tissue sample (e.g., liver, kidney, heart, muscle, adipose)",
                "name": "tissue_type",
                "type": "str",
            },
            {
                "default": None,
                "description": "Identifier for the sample being analyzed",
                "name": "sample_id",
                "type": "str",
            },
        ],
    },
    {
        "description": "Analyzes raw blood pressure data to calculate key hemodynamic parameters.",
        "name": "analyze_hemodynamic_data",
        "optional_parameters": [
            {
                "default": "hemodynamic_results.csv",
                "description": "Filename to save the calculated parameters",
                "name": "output_file",
                "type": "str",
            }
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Raw blood pressure measurements in mmHg",
                "name": "pressure_data",
                "type": "numpy.ndarray",
            },
            {
                "default": None,
                "description": "Data acquisition rate in Hz (samples per second)",
                "name": "sampling_rate",
                "type": "float",
            },
        ],
    },
    {
        "description": "Simulates the transport and binding of thyroid hormones "
        "across different tissue compartments using an ODE-based "
        "pharmacokinetic model.",
        "name": "simulate_thyroid_hormone_pharmacokinetics",
        "optional_parameters": [
            {
                "default": "(0, 24)",
                "description": "Start and end time for simulation in hours",
                "name": "time_span",
                "type": "tuple",
            },
            {
                "default": 100,
                "description": "Number of time points to output",
                "name": "time_points",
                "type": "int",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Dictionary containing model "
                "parameters including "
                "transport_rates, binding_constants, "
                "metabolism_rates, and volumes",
                "name": "parameters",
                "type": "dict",
            },
            {
                "default": None,
                "description": "Dictionary of initial concentrations for all molecular species in each compartment",
                "name": "initial_conditions",
                "type": "dict",
            },
        ],
    },
    {
        "description": "Analyzes an image to detect and quantify amyloid-beta "
        "plaques, returning a detailed analysis log.",
        "name": "quantify_amyloid_beta_plaques",
        "optional_parameters": [
            {
                "default": "./results",
                "description": "Directory where results will be saved",
                "name": "output_dir",
                "type": "str",
            },
            {
                "default": "otsu",
                "description": "Method for image thresholding (otsu, adaptive, or manual)",
                "name": "threshold_method",
                "type": "str",
            },
            {
                "default": 50,
                "description": "Minimum size in pixels² for a region to be considered a plaque",
                "name": "min_plaque_size",
                "type": "int",
            },
            {
                "default": 127,
                "description": "Threshold value to use when threshold_method is manual",
                "name": "manual_threshold",
                "type": "int",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the image file to be analyzed for amyloid-beta plaques",
                "name": "image_path",
                "type": "str",
            }
        ],
    },
]



def query_chatnt(question, sequence, device=-1):
    """
    Call ChatNT to answer a question about a DNA sequence.

    Parameters:
    -----------
    question : str
        Question to ask about the DNA sequence
    sequence : str
        DNA sequence to analyze
    device : int, optional
        Device to use for the ChatNT model. Default is -1 (CPU).

    Returns:
    --------
    str
        Answer to the question
    """
    import torch
    from transformers import pipeline

    pipe = pipeline(model="InstaDeepAI/ChatNT", trust_remote_code=True, device=device)

    # Define custom inputs (note that the number of <DNA> token in the english sequence must be equal to len(dna_sequences))
    english_sequence = f"{question} <DNA> ?"
    dna_sequences = [sequence]

    # Generate sequence
    generated_english_sequence = pipe(inputs={"english_sequence": english_sequence, "dna_sequences": dna_sequences})

    return generated_english_sequence


def perform_flux_balance_analysis(model_file, constraints=None, objective_reaction=None, output_file="fba_results.csv"):
    """
    Perform Flux Balance Analysis (FBA) on a genome-scale metabolic network model.

    FBA is a computational technique that predicts metabolic flux distributions
    by formulating and solving a linear optimization problem.

    Parameters:
    -----------
    model_file : str
        Path to the metabolic model file (SBML or JSON format)
    constraints : dict, optional
        Dictionary of reaction constraints where keys are reaction IDs and
        values are tuples of (lower_bound, upper_bound)
    objective_reaction : str, optional
        Reaction ID to use as the objective function (e.g., biomass reaction)
        If None, uses the model's default objective
    output_file : str, optional
        File name to save the flux distribution results

    Returns:
    --------
    str
        Research log summarizing the FBA process and results
    """
    import os

    import cobra
    import pandas as pd

    # Initialize research log
    log = "# Flux Balance Analysis (FBA) Research Log\n\n"

    # Step 1: Load the metabolic network model
    log += "## Step 1: Loading metabolic model\n"
    try:
        if model_file.endswith(".xml") or model_file.endswith(".sbml"):
            model = cobra.io.read_sbml_model(model_file)
        elif model_file.endswith(".json"):
            model = cobra.io.load_json_model(model_file)
        else:
            model = cobra.io.load_model(model_file)
        log += f"- Successfully loaded model from {model_file}\n"
        log += f"- Model contains {len(model.reactions)} reactions and {len(model.metabolites)} metabolites\n\n"
    except Exception as e:
        log += f"- Error loading model: {str(e)}\n"
        return log

    # Step 2: Set constraints
    log += "## Step 2: Setting constraints\n"
    if constraints:
        log += "- Applied the following constraints:\n"
        for reaction_id, (lb, ub) in constraints.items():
            try:
                reaction = model.reactions.get_by_id(reaction_id)
                reaction.bounds = (lb, ub)
                log += f"  * {reaction_id}: lower_bound={lb}, upper_bound={ub}\n"
            except Exception as e:
                log += f"  * Error setting constraint for {reaction_id}: {str(e)}\n"
    else:
        log += "- No additional constraints specified, using model defaults\n"
    log += "\n"

    # Step 3: Set objective function
    log += "## Step 3: Setting objective function\n"
    if objective_reaction:
        try:
            model.objective = objective_reaction
            log += f"- Set objective function to maximize {objective_reaction}\n\n"
        except Exception as e:
            log += f"- Error setting objective function: {str(e)}\n"
            log += "- Using model's default objective function\n\n"
    else:
        log += f"- Using model's default objective function: {model.objective.expression}\n\n"

    # Step 4: Solve the FBA problem
    log += "## Step 4: Solving FBA optimization problem\n"
    try:
        solution = model.optimize()
        log += f"- Optimization status: {solution.status}\n"
        log += f"- Objective value: {solution.objective_value:.6f}\n\n"
    except Exception as e:
        log += f"- Error during optimization: {str(e)}\n"
        return log

    # Step 5: Save and report results
    log += "## Step 5: Analyzing flux distribution\n"

    # Create a dataframe with the flux distribution
    flux_distribution = pd.DataFrame(
        {
            "reaction_id": [r.id for r in model.reactions],
            "reaction_name": [r.name for r in model.reactions],
            "flux": [solution.fluxes[r.id] for r in model.reactions],
            "lower_bound": [r.lower_bound for r in model.reactions],
            "upper_bound": [r.upper_bound for r in model.reactions],
        }
    )

    # Save to file
    flux_distribution.to_csv(output_file, index=False)
    log += f"- Flux distribution saved to {output_file}\n"

    # Report active reactions (non-zero flux)
    active_reactions = flux_distribution[abs(flux_distribution["flux"]) > 1e-6]
    log += f"- Number of active reactions (flux > 1e-6): {len(active_reactions)}\n"

    # Report top reactions by absolute flux
    top_reactions = flux_distribution.iloc[abs(flux_distribution["flux"]).argsort()[::-1]].head(10)
    log += "- Top 10 reactions by absolute flux magnitude:\n"
    for _, row in top_reactions.iterrows():
        log += f"  * {row['reaction_id']} ({row['reaction_name']}): {row['flux']:.6f}\n"

    log += f"\nFBA analysis complete. Full results available in {output_file}\n"

    return log


def model_protein_dimerization_network(monomer_concentrations, dimerization_affinities, network_topology):
    """
    Model protein dimerization networks to find equilibrium concentrations of dimers.

    Parameters:
    -----------
    monomer_concentrations : dict
        Dictionary mapping monomer names to their initial concentrations (in arbitrary units)
    dimerization_affinities : dict
        Dictionary mapping dimer names (as 'A-B' strings) to their association constants (Ka)
    network_topology : list of tuples
        List of (monomer1, monomer2) pairs that can form dimers

    Returns:
    --------
    str
        Research log summarizing the modeling process and results
    """
    import time

    import numpy as np
    from scipy.integrate import solve_ivp

    # Start timing
    start_time = time.time()

    # Extract unique monomers and create mapping to indices
    all_monomers = list(monomer_concentrations.keys())
    monomer_to_idx = {monomer: i for i, monomer in enumerate(all_monomers)}

    # Create mapping from dimer indices to names
    dimer_pairs = []
    dimer_names = []
    dimer_affinities = []

    for m1, m2 in network_topology:
        dimer_name = f"{m1}-{m2}"
        if dimer_name in dimerization_affinities:
            affinity = dimerization_affinities[dimer_name]
        else:
            # Try reverse order
            dimer_name = f"{m2}-{m1}"
            if dimer_name in dimerization_affinities:
                affinity = dimerization_affinities[dimer_name]
            else:
                raise ValueError(f"No affinity found for dimer pair {m1}-{m2}")

        dimer_pairs.append((monomer_to_idx[m1], monomer_to_idx[m2]))
        dimer_names.append(dimer_name)
        dimer_affinities.append(affinity)

    # Initial conditions (monomer concentrations followed by dimer concentrations)
    num_monomers = len(all_monomers)
    num_dimers = len(dimer_pairs)
    y0 = np.zeros(num_monomers + num_dimers)

    for monomer, conc in monomer_concentrations.items():
        y0[monomer_to_idx[monomer]] = conc

    # Define ODE system
    def dimerization_odes(t, y):
        dydt = np.zeros_like(y)

        # Extract current concentrations
        monomer_concs = y[:num_monomers]
        dimer_concs = y[num_monomers:]

        # Calculate rate of change for each species
        for i, ((m1_idx, m2_idx), affinity) in enumerate(zip(dimer_pairs, dimer_affinities, strict=False)):
            # Formation rate: kon * [A] * [B]
            # Dissociation rate: koff * [AB]
            # At equilibrium: kon/koff = Ka (affinity)
            # For simplicity, set kon = Ka and koff = 1

            kon = affinity
            koff = 1.0

            # Dimer formation/dissociation
            dimer_idx = num_monomers + i

            # Rate of change
            formation_rate = kon * monomer_concs[m1_idx] * monomer_concs[m2_idx]
            dissociation_rate = koff * dimer_concs[i]
            net_rate = formation_rate - dissociation_rate

            # Update rates for monomers and dimers
            dydt[m1_idx] -= net_rate
            dydt[m2_idx] -= net_rate
            dydt[dimer_idx] += net_rate

        return dydt

    # Solve ODE system to equilibrium
    # Using a long enough time to reach equilibrium
    t_span = (0, 1000)  # Arbitrary long time to reach equilibrium

    sol = solve_ivp(
        dimerization_odes,
        t_span,
        y0,
        method="BDF",  # Good for stiff problems
        rtol=1e-6,
        atol=1e-9,
    )

    # Extract final concentrations (equilibrium)
    final_monomer_concs = {monomer: sol.y[idx][-1] for monomer, idx in monomer_to_idx.items()}
    final_dimer_concs = {dimer_name: sol.y[num_monomers + i][-1] for i, dimer_name in enumerate(dimer_names)}

    # Calculate time taken
    elapsed_time = time.time() - start_time

    # Generate research log
    log = "Protein Dimerization Network Modeling\n"
    log += "=====================================\n\n"
    log += "Network summary:\n"
    log += f"- Number of monomers: {num_monomers}\n"
    log += f"- Number of possible dimers: {num_dimers}\n\n"

    log += "Initial conditions:\n"
    for monomer, conc in monomer_concentrations.items():
        log += f"- {monomer}: {conc:.4f}\n"
    log += "\n"

    log += "Dimerization affinities:\n"
    for dimer, affinity in zip(dimer_names, dimer_affinities, strict=False):
        log += f"- {dimer}: {affinity:.4e}\n"
    log += "\n"

    log += "Equilibrium concentrations:\n"
    log += "Monomers:\n"
    for monomer, conc in final_monomer_concs.items():
        log += f"- {monomer}: {conc:.4f}\n"

    log += "\nDimers:\n"
    for dimer, conc in final_dimer_concs.items():
        log += f"- {dimer}: {conc:.4f}\n"
    log += "\n"

    log += f"Simulation completed in {elapsed_time:.2f} seconds.\n"

    return log


def simulate_metabolic_network_perturbation(
    model_file, initial_concentrations, perturbation_params, simulation_time=100, time_points=1000
):
    """
    Construct and simulate kinetic models of metabolic networks and analyze their responses to perturbations.

    Parameters:
    -----------
    model_file : str
        Path to the COBRA model file (SBML format)
    initial_concentrations : dict
        Dictionary mapping metabolite IDs to their initial concentrations
    perturbation_params : dict
        Dictionary with the following keys:
        - 'time': float, time at which perturbation is applied
        - 'metabolite': str, ID of the metabolite to perturb
        - 'factor': float, multiplication factor for the metabolite concentration
    simulation_time : float, optional
        Total simulation time (default: 100)
    time_points : int, optional
        Number of time points to simulate (default: 1000)

    Returns:
    --------
    str
        Research log summarizing the steps taken and results obtained
    """
    import os

    import cobra
    import numpy as np
    import pandas as pd
    from scipy.integrate import solve_ivp

    # Step 1: Load the metabolic model
    try:
        model = cobra.io.read_sbml_model(model_file)
        log = f"Loaded metabolic model from {model_file} with {len(model.reactions)} reactions and {len(model.metabolites)} metabolites.\n\n"
    except Exception as e:
        return f"Error loading model: {str(e)}"

    # Step 2: Set up metabolite list and initial concentrations
    metabolites = list(model.metabolites)
    metabolite_ids = [m.id for m in metabolites]

    # Set default initial concentrations for metabolites not specified
    conc_array = np.ones(len(metabolites))
    for i, m_id in enumerate(metabolite_ids):
        if m_id in initial_concentrations:
            conc_array[i] = initial_concentrations[m_id]

    log += "Initial concentrations set up for all metabolites.\n\n"

    # Step 3: Define kinetic model using simple mass-action kinetics
    def mass_action_kinetics(reaction, concentrations, metabolite_ids):
        # Simple mass-action kinetics: product of substrate concentrations
        rate = 1.0
        for metabolite, coeff in reaction.metabolites.items():
            if coeff < 0:  # Substrate
                idx = metabolite_ids.index(metabolite.id)
                rate *= concentrations[idx] ** abs(coeff)
        return rate

    # Step 4: Define ODE system
    def ode_system(t, concentrations, metabolite_ids, reactions, perturbation):
        dC_dt = np.zeros(len(concentrations))

        # Apply perturbation if we're at or past the perturbation time
        if t >= perturbation["time"]:
            perturb_idx = metabolite_ids.index(perturbation["metabolite"])
            if not perturbation.get("applied", False):
                concentrations[perturb_idx] *= perturbation["factor"]
                perturbation["applied"] = True

        # Calculate reaction rates
        rates = {}
        for reaction in reactions:
            rates[reaction.id] = mass_action_kinetics(reaction, concentrations, metabolite_ids)

        # Update concentration changes based on stoichiometry and rates
        for i, metabolite_id in enumerate(metabolite_ids):
            for reaction in reactions:
                if metabolite_id in [m.id for m in reaction.metabolites]:
                    coeff = reaction.metabolites[model.metabolites.get_by_id(metabolite_id)]
                    dC_dt[i] += coeff * rates[reaction.id]

        return dC_dt

    log += "Kinetic model defined using mass-action kinetics for all reactions.\n\n"

    # Step 5: Simulate the system
    perturbation = perturbation_params.copy()
    perturbation["applied"] = False

    t_span = (0, simulation_time)
    t_eval = np.linspace(0, simulation_time, time_points)

    log += f"Starting simulation for {simulation_time} time units with perturbation of {perturbation['metabolite']} "
    log += f"by factor {perturbation['factor']} at time {perturbation['time']}.\n\n"

    try:
        solution = solve_ivp(
            lambda t, y: ode_system(t, y, metabolite_ids, model.reactions, perturbation),
            t_span,
            conc_array,
            t_eval=t_eval,
            method="LSODA",
        )

        log += f"Simulation completed successfully with {len(solution.t)} time points.\n\n"
    except Exception as e:
        return f"Error during simulation: {str(e)}"

    # Step 6: Calculate fluxes at each time point
    fluxes = np.zeros((len(solution.t), len(model.reactions)))
    reaction_ids = [r.id for r in model.reactions]

    for t_idx, t in enumerate(solution.t):
        concentrations = solution.y[:, t_idx]
        perturbation_temp = perturbation.copy()
        if t >= perturbation["time"]:
            perturbation_temp["applied"] = True

        for r_idx, reaction in enumerate(model.reactions):
            fluxes[t_idx, r_idx] = mass_action_kinetics(reaction, concentrations, metabolite_ids)

    log += "Calculated reaction fluxes for all time points.\n\n"

    # Step 7: Save results to files
    conc_df = pd.DataFrame(solution.y.T, columns=metabolite_ids)
    conc_df["time"] = solution.t
    conc_file = "metabolite_concentrations.csv"
    conc_df.to_csv(conc_file, index=False)

    flux_df = pd.DataFrame(fluxes, columns=reaction_ids)
    flux_df["time"] = solution.t
    flux_file = "reaction_fluxes.csv"
    flux_df.to_csv(flux_file, index=False)

    log += f"Results saved to {conc_file} and {flux_file}.\n\n"

    # Step 8: Analyze perturbation response
    perturb_time_idx = np.where(solution.t >= perturbation["time"])[0][0]
    pre_perturb = solution.y[:, perturb_time_idx - 1]
    post_perturb = solution.y[:, perturb_time_idx]

    # Find metabolites with significant changes
    significant_changes = []
    for i, m_id in enumerate(metabolite_ids):
        rel_change = abs(post_perturb[i] - pre_perturb[i]) / pre_perturb[i] if pre_perturb[i] > 0 else 0
        if rel_change > 0.05:  # 5% change threshold
            significant_changes.append((m_id, rel_change))

    log += "Perturbation Analysis Results:\n"
    log += f"Perturbation of {perturbation['metabolite']} at time {perturbation['time']} by factor {perturbation['factor']}.\n"
    log += f"Number of metabolites with significant immediate response: {len(significant_changes)}.\n"

    if significant_changes:
        log += "Top 5 most affected metabolites (by relative concentration change):\n"
        for m_id, change in sorted(significant_changes, key=lambda x: x[1], reverse=True)[:5]:
            log += f"  - {m_id}: {change * 100:.2f}% change\n"

    log += "\nSimulation and perturbation analysis completed successfully."

    return log


def simulate_protein_signaling_network(
    network_structure, reaction_params, species_params, simulation_time=100, time_points=1000
):
    """
    Simulate protein signaling network dynamics using ODE-based logic modeling with normalized Hill functions.

    Parameters:
    -----------
    network_structure : dict
        Dictionary defining the network topology. Each key is a target protein and its value is a list of tuples
        (regulator, regulation_type) where regulation_type is 1 for activation and -1 for inhibition.

    reaction_params : dict
        Dictionary of reaction parameters. Keys are tuples (regulator, target) and values are dictionaries
        with keys 'W' (weight), 'n' (Hill coefficient), and 'EC50' (half-maximal effective concentration).

    species_params : dict
        Dictionary of species parameters. Keys are protein names and values are dictionaries
        with keys 'tau' (time constant), 'y0' (initial concentration), and 'ymax' (maximum concentration).

    simulation_time : float, optional
        Total simulation time in arbitrary time units. Default is 100.

    time_points : int, optional
        Number of time points for the simulation. Default is 1000.

    Returns:
    --------
    str
        Research log summarizing the simulation process and results.
    """
    import csv

    import numpy as np
    from scipy.integrate import solve_ivp

    # Extract all unique protein species
    all_proteins = set(network_structure.keys())
    for _target, regulators in network_structure.items():
        for regulator, _ in regulators:
            all_proteins.add(regulator)

    # Create ordered list of proteins for indexing
    proteins = sorted(all_proteins)
    protein_indices = {protein: i for i, protein in enumerate(proteins)}

    # Define the normalized Hill function
    def hill_function(x, n, ec50):
        return x**n / (x**n + ec50**n)

    # Define the ODE system
    def ode_system(t, y):
        dydt = np.zeros_like(y)

        for target, regulators in network_structure.items():
            target_idx = protein_indices[target]
            target_params = species_params[target]

            # Calculate regulation term for each regulator
            regulation_terms = []
            for regulator, reg_type in regulators:
                regulator_idx = protein_indices[regulator]
                params = reaction_params.get((regulator, target), {})

                if not params:
                    continue

                x = y[regulator_idx]
                n = params["n"]
                ec50 = params["EC50"]
                weight = params["W"]

                # Apply Hill function based on regulation type
                if reg_type == 1:  # Activation
                    term = weight * hill_function(x, n, ec50)
                else:  # Inhibition
                    term = weight * (1 - hill_function(x, n, ec50))

                regulation_terms.append(term)

            # Combine regulation terms (if any)
            if regulation_terms:
                # Simple summation of regulation terms
                regulation = sum(regulation_terms) / len(regulation_terms)

                # Ensure regulation stays within [0, 1]
                regulation = max(0, min(1, regulation))

                # Calculate rate of change
                tau = target_params["tau"]
                ymax = target_params["ymax"]
                dydt[target_idx] = (1 / tau) * (regulation * ymax - y[target_idx])

        return dydt

    # Set up initial conditions
    y0 = np.zeros(len(proteins))
    for protein, params in species_params.items():
        if protein in protein_indices:
            y0[protein_indices[protein]] = params["y0"]

    # Set up time points
    t_span = (0, simulation_time)
    t_eval = np.linspace(0, simulation_time, time_points)

    # Solve the ODE system
    solution = solve_ivp(ode_system, t_span, y0, method="LSODA", t_eval=t_eval, rtol=1e-6, atol=1e-9)

    # Save results to CSV
    output_file = "protein_signaling_simulation_results.csv"
    with open(output_file, "w", newline="") as f:
        writer = csv.writer(f)
        # Write header
        header = ["Time"] + proteins
        writer.writerow(header)

        # Write data
        for i in range(len(solution.t)):
            row = [solution.t[i]] + list(solution.y[:, i])
            writer.writerow(row)

    # Generate research log
    log = "ODE-based Logic Modeling of Protein Signaling Networks\n"
    log += "=============================================\n\n"
    log += f"Network structure: {len(network_structure)} target proteins, {len(proteins)} total proteins\n"
    log += f"Simulation time: {simulation_time} time units\n"
    log += f"Number of time points: {time_points}\n\n"

    log += "Simulation steps:\n"
    log += "1. Initialized protein concentrations based on provided initial values\n"
    log += "2. Set up ODE system using normalized Hill functions for protein interactions\n"
    log += "3. Solved the ODE system using LSODA integration method\n"
    log += "4. Saved time-series data for all protein concentrations\n\n"

    log += f"Results saved to: {output_file}\n\n"

    # Add summary statistics
    log += "Summary of final protein concentrations:\n"
    for i, protein in enumerate(proteins):
        final_conc = solution.y[i, -1]
        log += f"- {protein}: {final_conc:.4f}\n"

    return log


def compare_protein_structures(pdb_file1, pdb_file2, chain_id1="A", chain_id2="A", output_prefix="protein_comparison"):
    """
    Compares two protein structures to identify structural differences and conformational changes.

    Parameters:
    -----------
    pdb_file1 : str
        Path to the first PDB file
    pdb_file2 : str
        Path to the second PDB file
    chain_id1 : str, optional
        Chain ID to analyze in the first structure (default: 'A')
    chain_id2 : str, optional
        Chain ID to analyze in the second structure (default: 'A')
    output_prefix : str, optional
        Prefix for output files (default: "protein_comparison")

    Returns:
    --------
    str
        A research log summarizing the structural comparison analysis
    """
    import os
    import warnings

    import numpy as np
    from Bio.PDB import PDBIO, PDBParser, Select, Superimposer
    from Bio.PDB.PDBExceptions import PDBConstructionWarning

    # Suppress PDB parsing warnings
    warnings.filterwarnings("ignore", category=PDBConstructionWarning)

    research_log = []
    research_log.append("# Protein Structure Comparison Analysis\n")
    research_log.append(f"Comparing structures from {pdb_file1} and {pdb_file2}\n")

    # Initialize parser
    parser = PDBParser()

    # Parse structures
    research_log.append("## Loading PDB structures")
    structure1 = parser.get_structure("structure1", pdb_file1)
    structure2 = parser.get_structure("structure2", pdb_file2)

    # Get specified chains
    try:
        chain1 = structure1[0][chain_id1]
        chain2 = structure2[0][chain_id2]
        research_log.append(f"Successfully loaded chain {chain_id1} from {pdb_file1}")
        research_log.append(f"Successfully loaded chain {chain_id2} from {pdb_file2}")
    except KeyError as e:
        return f"Error: Chain not found: {str(e)}"

    # Get CA atoms for alignment
    ca_atoms1 = []
    ca_atoms2 = []

    # Create residue mappings based on residue ID
    residues1 = {residue.id[1]: residue for residue in chain1 if residue.has_id("CA")}
    residues2 = {residue.id[1]: residue for residue in chain2 if residue.has_id("CA")}

    # Find common residues
    common_residues = sorted(set(residues1.keys()) & set(residues2.keys()))

    if len(common_residues) == 0:
        return "Error: No common residues with CA atoms found between the structures"

    research_log.append(f"Found {len(common_residues)} common residues for alignment\n")

    # Get paired atoms for alignment
    for res_id in common_residues:
        ca_atoms1.append(residues1[res_id]["CA"])
        ca_atoms2.append(residues2[res_id]["CA"])

    # Perform structural alignment
    research_log.append("## Structural Alignment")
    super_imposer = Superimposer()
    super_imposer.set_atoms(ca_atoms1, ca_atoms2)
    super_imposer.apply(structure2.get_atoms())

    # Calculate RMSD
    rmsd = super_imposer.rms
    research_log.append(f"Overall RMSD: {rmsd:.4f} Å\n")

    # Calculate per-residue distance after alignment
    research_log.append("## Conformational Differences Analysis")

    # Find regions with significant differences
    distance_data = []
    significant_changes = []
    threshold = 2.0  # Angstroms threshold for significant difference

    for res_id in common_residues:
        res1 = residues1[res_id]
        res2 = residues2[res_id]

        # Calculate distance between CA atoms
        ca1 = res1["CA"]
        ca2 = res2["CA"]
        distance = np.linalg.norm(ca1.coord - ca2.coord)

        distance_data.append((res_id, distance))

        if distance > threshold:
            significant_changes.append((res_id, res1.get_resname(), distance))

    # Save alignment as PDB files
    aligned_file1 = f"{output_prefix}_ref.pdb"
    aligned_file2 = f"{output_prefix}_aligned.pdb"

    io = PDBIO()
    io.set_structure(structure1)
    io.save(aligned_file1, select=Select())

    io.set_structure(structure2)
    io.save(aligned_file2, select=Select())

    research_log.append(f"Aligned structures saved as {aligned_file1} and {aligned_file2}")

    # Report on significant differences
    if significant_changes:
        research_log.append(
            f"\nIdentified {len(significant_changes)} residues with significant conformational changes:"
        )
        for res_id, res_name, distance in significant_changes:
            research_log.append(f"  - Residue {res_name}{res_id}: {distance:.2f} Å displacement")
    else:
        research_log.append("\nNo significant conformational changes detected (threshold: 2.0 Å)")

    # Save per-residue distance data
    distance_file = f"{output_prefix}_residue_distances.csv"
    with open(distance_file, "w") as f:
        f.write("Residue_ID,Distance(Å)\n")
        for res_id, distance in distance_data:
            f.write(f"{res_id},{distance:.4f}\n")

    research_log.append(f"\nPer-residue distance data saved to {distance_file}")

    # Identify regions with continuous changes
    regions = []
    current_region = []

    for res_id, _, distance in significant_changes:
        if not current_region or res_id == current_region[-1][0] + 1:
            current_region.append((res_id, distance))
        else:
            if len(current_region) >= 3:  # Consider regions with at least 3 consecutive residues
                regions.append(current_region)
            current_region = [(res_id, distance)]

    if current_region and len(current_region) >= 3:
        regions.append(current_region)

    if regions:
        research_log.append("\n## Continuous Regions with Conformational Changes")
        for i, region in enumerate(regions, 1):
            start_res = region[0][0]
            end_res = region[-1][0]
            avg_dist = sum(r[1] for r in region) / len(region)
            research_log.append(f"Region {i}: Residues {start_res}-{end_res} (Average displacement: {avg_dist:.2f} Å)")

    research_log.append("\n## Summary")
    research_log.append(f"- Compared structures from {pdb_file1} and {pdb_file2}")
    research_log.append(f"- Overall RMSD: {rmsd:.4f} Å")
    research_log.append(f"- {len(significant_changes)} residues with significant conformational changes")
    research_log.append(f"- {len(regions)} continuous regions of conformational change")
    research_log.append(f"- Files generated: {aligned_file1}, {aligned_file2}, {distance_file}")

    return "\n".join(research_log)


def simulate_renin_angiotensin_system_dynamics(
    initial_concentrations, rate_constants, feedback_params, simulation_time=48, time_points=100
):
    """
    Simulate the time-dependent concentrations of renin-angiotensin system (RAS) components.

    Parameters:
    -----------
    initial_concentrations : dict
        Initial concentrations of RAS components with keys:
        'renin', 'angiotensinogen', 'angiotensin_I', 'angiotensin_II',
        'ACE2_angiotensin_II', 'angiotensin_1_7'

    rate_constants : dict
        Kinetic rate constants with keys:
        'k_ren' (renin production), 'k_agt' (angiotensinogen production),
        'k_ace' (ACE conversion rate), 'k_ace2' (ACE2 conversion rate),
        'k_at1r' (AT1R binding rate), 'k_mas' (Mas receptor binding rate)

    feedback_params : dict
        Parameters controlling feedback mechanisms with keys:
        'fb_ang_II' (angiotensin II feedback), 'fb_ace2' (ACE2 feedback)

    simulation_time : float, optional
        Total simulation time in hours (default: 48)

    time_points : int, optional
        Number of time points to evaluate (default: 100)

    Returns:
    --------
    str
        Research log summarizing the simulation steps and results
    """
    import os

    import numpy as np
    import pandas as pd
    from scipy.integrate import solve_ivp

    # Extract initial concentrations
    y0 = [
        initial_concentrations["renin"],
        initial_concentrations["angiotensinogen"],
        initial_concentrations["angiotensin_I"],
        initial_concentrations["angiotensin_II"],
        initial_concentrations["ACE2_angiotensin_II"],
        initial_concentrations["angiotensin_1_7"],
    ]

    # Define the system of ODEs
    def ras_ode_system(t, y):
        renin, agt, ang_I, ang_II, ace2_ang_II, ang_1_7 = y

        # Production rates with feedback
        renin_production = rate_constants["k_ren"] * (1 / (1 + feedback_params["fb_ang_II"] * ang_II))
        agt_production = rate_constants["k_agt"]

        # Conversion rates
        ang_I_formation = renin * agt
        ang_II_formation = rate_constants["k_ace"] * ang_I
        ace2_binding = rate_constants["k_ace2"] * ang_II
        ang_1_7_formation = ace2_ang_II

        # Clearance/degradation (simplified as proportional to concentration)
        renin_clearance = 0.1 * renin
        agt_clearance = 0.05 * agt
        ang_I_clearance = 0.2 * ang_I
        ang_II_clearance = 0.3 * ang_II + rate_constants["k_at1r"] * ang_II
        ace2_ang_II_clearance = 0.15 * ace2_ang_II
        ang_1_7_clearance = 0.25 * ang_1_7 + rate_constants["k_mas"] * ang_1_7

        # ODEs
        drenin_dt = renin_production - renin_clearance
        dagt_dt = agt_production - agt_clearance - ang_I_formation
        dang_I_dt = ang_I_formation - ang_I_clearance - ang_II_formation
        dang_II_dt = ang_II_formation - ang_II_clearance - ace2_binding
        dace2_ang_II_dt = ace2_binding - ace2_ang_II_clearance - ang_1_7_formation
        dang_1_7_dt = ang_1_7_formation - ang_1_7_clearance

        return [drenin_dt, dagt_dt, dang_I_dt, dang_II_dt, dace2_ang_II_dt, dang_1_7_dt]

    # Time points for simulation
    t_span = (0, simulation_time)
    t_eval = np.linspace(0, simulation_time, time_points)

    # Solve the ODE system
    solution = solve_ivp(ras_ode_system, t_span, y0, method="RK45", t_eval=t_eval, rtol=1e-6)

    # Create DataFrame with results
    component_names = [
        "Renin",
        "Angiotensinogen",
        "Angiotensin I",
        "Angiotensin II",
        "ACE2-Angiotensin II",
        "Angiotensin 1-7",
    ]
    results_df = pd.DataFrame(solution.y.T, columns=component_names)
    results_df.insert(0, "Time (hours)", solution.t)

    # Save results to CSV
    results_file = "ras_simulation_results.csv"
    results_df.to_csv(results_file, index=False)

    # Create research log
    log = f"""RAS ODE Modeling Simulation Log:

1. Initialized RAS component concentrations:
   - Renin: {initial_concentrations["renin"]}
   - Angiotensinogen: {initial_concentrations["angiotensinogen"]}
   - Angiotensin I: {initial_concentrations["angiotensin_I"]}
   - Angiotensin II: {initial_concentrations["angiotensin_II"]}
   - ACE2-Angiotensin II complex: {initial_concentrations["ACE2_angiotensin_II"]}
   - Angiotensin 1-7: {initial_concentrations["angiotensin_1_7"]}

2. Applied rate constants:
   - Renin production (k_ren): {rate_constants["k_ren"]}
   - Angiotensinogen production (k_agt): {rate_constants["k_agt"]}
   - ACE conversion rate (k_ace): {rate_constants["k_ace"]}
   - ACE2 conversion rate (k_ace2): {rate_constants["k_ace2"]}
   - AT1R binding rate (k_at1r): {rate_constants["k_at1r"]}
   - Mas receptor binding rate (k_mas): {rate_constants["k_mas"]}

3. Feedback parameters:
   - Angiotensin II feedback (fb_ang_II): {feedback_params["fb_ang_II"]}
   - ACE2 feedback (fb_ace2): {feedback_params["fb_ace2"]}

4. Simulation parameters:
   - Total simulation time: {simulation_time} hours
   - Number of time points: {time_points}

5. Solved system of ODEs using SciPy's solve_ivp with RK45 method.

6. Final concentrations at {simulation_time} hours:
   - Renin: {results_df["Renin"].iloc[-1]:.4f}
   - Angiotensinogen: {results_df["Angiotensinogen"].iloc[-1]:.4f}
   - Angiotensin I: {results_df["Angiotensin I"].iloc[-1]:.4f}
   - Angiotensin II: {results_df["Angiotensin II"].iloc[-1]:.4f}
   - ACE2-Angiotensin II complex: {results_df["ACE2-Angiotensin II"].iloc[-1]:.4f}
   - Angiotensin 1-7: {results_df["Angiotensin 1-7"].iloc[-1]:.4f}

7. Results saved to file: {results_file}

8. Key observations:
   - The model captures the conversion of Angiotensinogen to Angiotensin I via Renin
   - Angiotensin I is converted to Angiotensin II via ACE
   - Angiotensin II binds with ACE2 to form the ACE2-Angiotensin II complex
   - The complex is converted to Angiotensin 1-7
   - Feedback mechanisms regulate the production rates
"""

    return log



description = [
    {
        "description": "Perform ATAC-seq peak calling and differential accessibility analysis using MACS2.",
        "name": "analyze_atac_seq_differential_accessibility",
        "optional_parameters": [
            {
                "default": "./atac_results",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            },
            {
                "default": "hs",
                "description": "Genome size parameter for MACS2",
                "name": "genome_size",
                "type": "str",
            },
            {
                "default": 0.05,
                "description": "q-value cutoff for peak detection",
                "name": "q_value",
                "type": "float",
            },
            {
                "default": "atac",
                "description": "Prefix for output file names",
                "name": "name_prefix",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the treatment condition BAM file with aligned ATAC-seq reads",
                "name": "treatment_bam",
                "type": "str",
            },
            {
                "default": None,
                "description": "Path to the control condition BAM file with aligned ATAC-seq reads",
                "name": "control_bam",
                "type": "str",
            },
        ],
    },
    {
        "description": "Analyzes bacterial growth curve data to determine growth "
        "parameters such as doubling time, growth rate, and lag "
        "phase.",
        "name": "analyze_bacterial_growth_curve",
        "optional_parameters": [
            {
                "default": ".",
                "description": "Directory where output files will be saved",
                "name": "output_dir",
                "type": "str",
            }
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Time points of measurements in hours",
                "name": "time_points",
                "type": "List or numpy.ndarray",
            },
            {
                "default": None,
                "description": "Optical density measurements corresponding to each time point",
                "name": "od_values",
                "type": "List or numpy.ndarray",
            },
            {
                "default": None,
                "description": "Name of the bacterial strain being analyzed",
                "name": "strain_name",
                "type": "str",
            },
        ],
    },
    {
        "description": "Simulates the isolation and purification of immune cells from tissue samples.",
        "name": "isolate_purify_immune_cells",
        "optional_parameters": [
            {
                "default": "collagenase",
                "description": "The enzyme used for tissue digestion",
                "name": "enzyme_type",
                "type": "str",
            },
            {
                "default": None,
                "description": "Specific antibody for magnetic-assisted cell sorting",
                "name": "macs_antibody",
                "type": "str",
            },
            {
                "default": 45,
                "description": "Digestion time in minutes",
                "name": "digestion_time_min",
                "type": "int",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "The type of tissue sample (e.g., 'adipose', 'kidney', 'liver', 'lung', 'spleen')",
                "name": "tissue_type",
                "type": "str",
            },
            {
                "default": None,
                "description": "The immune cell population to isolate (e.g., 'macrophages', 'leukocytes', 'T cells')",
                "name": "target_cell_type",
                "type": "str",
            },
        ],
    },
    {
        "description": "Estimate cell cycle phase durations using dual-nucleoside "
        "pulse labeling data and mathematical modeling.",
        "name": "estimate_cell_cycle_phase_durations",
        "optional_parameters": [],
        "required_parameters": [
            {
                "default": None,
                "description": "Dictionary containing experimental "
                "data from flow cytometry with EdU "
                "and BrdU labeling, including time "
                "points and percentages of labeled "
                "cells",
                "name": "flow_cytometry_data",
                "type": "dict",
            },
            {
                "default": None,
                "description": "Initial estimates for cell cycle phase durations and death rates",
                "name": "initial_estimates",
                "type": "dict",
            },
        ],
    },
    {
        "description": "Track immune cells under flow conditions and classify their behaviors.",
        "name": "track_immune_cells_under_flow",
        "optional_parameters": [
            {
                "default": "./output",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            },
            {
                "default": 1.0,
                "description": "Pixel size in micrometers",
                "name": "pixel_size_um",
                "type": "float",
            },
            {
                "default": 1.0,
                "description": "Time interval between frames in seconds",
                "name": "time_interval_sec",
                "type": "float",
            },
            {
                "default": "right",
                "description": "Direction of flow ('right', 'left', 'up', 'down')",
                "name": "flow_direction",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to image sequence directory or video file",
                "name": "image_sequence_path",
                "type": "str",
            }
        ],
    },
    {
        "description": "Analyze CFSE-labeled cell samples to quantify cell division and proliferation.",
        "name": "analyze_cfse_cell_proliferation",
        "optional_parameters": [
            {
                "default": "FL1-A",
                "description": "Name of the channel containing CFSE fluorescence data",
                "name": "cfse_channel",
                "type": "str",
            },
            {
                "default": None,
                "description": "Tuple of (min_fsc, max_fsc, min_ssc, max_ssc) for lymphocyte gating",
                "name": "lymphocyte_gate",
                "type": "tuple or None",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the FCS file containing flow cytometry data from CFSE-labeled cells",
                "name": "fcs_file_path",
                "type": "str",
            }
        ],
    },
    {
        "description": "Analyze cytokine production (IFN-γ, IL-17) in CD4+ T cells after antigen stimulation.",
        "name": "analyze_cytokine_production_in_cd4_tcells",
        "optional_parameters": [
            {
                "default": "./results",
                "description": "Directory to save the results file",
                "name": "output_dir",
                "type": "str",
            }
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Dictionary mapping stimulation "
                "conditions to FCS file paths. "
                "Expected keys: 'unstimulated', "
                "'Mtb300', 'CMV', 'SEB'",
                "name": "fcs_files_dict",
                "type": "dict",
            }
        ],
    },
    {
        "description": "Analyze ELISA data to quantify EBV antibody titers in plasma/serum samples.",
        "name": "analyze_ebv_antibody_titers",
        "optional_parameters": [
            {
                "default": "./",
                "description": "Directory to save output files.",
                "name": "output_dir",
                "type": "str",
            }
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Dictionary containing optical "
                "density (OD) readings for each "
                "sample. Format: {sample_id: "
                "{'VCA_IgG': float, 'VCA_IgM': "
                "float, 'EA_IgG': float, 'EA_IgM': "
                "float, 'EBNA1_IgG': float, "
                "'EBNA1_IgM': float}}",
                "name": "raw_od_data",
                "type": "dict",
            },
            {
                "default": None,
                "description": "Dictionary containing standard "
                "curve data for each antibody type. "
                "Format: {antibody_type: "
                "[(concentration, OD), ...]}",
                "name": "standard_curve_data",
                "type": "dict",
            },
            {
                "default": None,
                "description": "Dictionary containing metadata for "
                "each sample. Format: {sample_id: "
                "{'group': str, 'collection_date': "
                "str}}",
                "name": "sample_metadata",
                "type": "dict",
            },
        ],
    },
    {
        "description": "Analyzes histological images of CNS lesions to quantify "
        "immune cell infiltration, demyelination, and tissue damage.",
        "name": "analyze_cns_lesion_histology",
        "optional_parameters": [
            {
                "default": "./output",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            },
            {
                "default": "H&E",
                "description": 'Type of histological stain used (options: "H&E", "LFB", "IHC")',
                "name": "stain_type",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the microscopy image file of brain or spinal cord tissue section",
                "name": "image_path",
                "type": "str",
            }
        ],
    },
    {
        "description": "Analyzes immunohistochemistry images to quantify protein expression and spatial distribution.",
        "name": "analyze_immunohistochemistry_image",
        "optional_parameters": [
            {
                "default": "Unknown",
                "description": "Name of the protein being analyzed",
                "name": "protein_name",
                "type": "str",
            },
            {
                "default": "./ihc_results/",
                "description": "Directory to save output files",
                "name": "output_dir",
                "type": "str",
            },
        ],
        "required_parameters": [
            {
                "default": None,
                "description": "Path to the microscopy image of tissue section stained with antibodies",
                "name": "image_path",
                "type": "str",
            }
        ],
    },
]



#!/usr/bin/env python3
"""Script to process all bioRxiv subjects and extract computational tasks from the top 100 papers in each."""

import argparse
import os
import subprocess
import sys
import time

import pandas as pd
from tqdm import tqdm

# Add the parent directory to the path so we can import the bioagentos package
sys.path.append("/dfs/user/kexinh/BioAgentOS/")

# List of all bioRxiv subjects
BIORXIV_SUBJECTS = [
    "evolutionary biology",
    "ecology",
    "neuroscience",
    "developmental biology",
    "plant biology",
    "microbiology",
    "cancer biology",
    "immunology",
    "cell biology",
    "biochemistry",
    "genetics",
    "bioinformatics",
    "animal behavior and cognition",
    "biophysics",
    "genomics",
    "systems biology",
    "bioengineering",
    "molecular biology",
    "physiology",
    "zoology",
    "scientific communication and education",
    "pathology",
    "synthetic biology",
    "paleontology",
    "pharmacology and toxicology",
]


def parse_arguments():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description="Process all bioRxiv subjects and extract tasks from papers.")
    parser.add_argument(
        "--papers-per-subject",
        type=int,
        default=100,
        help="Number of papers to process per subject (default: 100)",
    )
    parser.add_argument(
        "--metadata-path",
        type=str,
        default="/dfs/user/kexinh/BioAgentOS/data/biorxiv_metadata.csv",
        help="Path to bioRxiv metadata CSV file (default: data/biorxiv_metadata.csv)",
    )
    parser.add_argument(
        "--output-dir",
        type=str,
        default="./biorxiv_all_subjects",
        help="Base directory to save results (default: ./biorxiv_all_subjects)",
    )
    parser.add_argument(
        "--model",
        type=str,
        default="claude-3-haiku-20240307",
        help="LLM model to use for extraction (default: claude-3-haiku-20240307)",
    )
    parser.add_argument(
        "--max-paper-length",
        type=int,
        default=200000,
        help="Maximum paper length in characters (default: 200000)",
    )
    parser.add_argument(
        "--chunk-size",
        type=int,
        default=10000,
        help="Chunk size for text processing (default: 10000)",
    )
    parser.add_argument(
        "--chunk-overlap",
        type=int,
        default=400,
        help="Chunk overlap for text processing (default: 400)",
    )
    parser.add_argument("--save-pdfs", action="store_true", help="Save downloaded PDFs (default: False)")
    parser.add_argument(
        "--random-sample",
        action="store_true",
        help="Randomly sample papers instead of taking the first N (default: False)",
    )
    parser.add_argument(
        "--subjects",
        type=str,
        nargs="+",
        help="Specific subjects to process (default: all subjects)",
    )
    parser.add_argument(
        "--delay",
        type=int,
        default=60,
        help="Delay in seconds between processing subjects (default: 60)",
    )
    parser.add_argument(
        "--summary-only",
        action="store_true",
        help="Only generate the combined summary (default: False)",
    )
    return parser.parse_args()


def check_subject_availability(metadata_path, subjects):
    """Check which subjects have papers available in the metadata.

    Args:
        metadata_path: Path to the metadata CSV file
        subjects: List of subjects to check

    Returns:
        List of subjects with available papers

    """
    try:
        df_biorxiv = pd.read_csv(metadata_path)
        available_subjects = []

        for subject in subjects:
            count = len(
                df_biorxiv[(df_biorxiv.published != "NA") & (df_biorxiv.category.str.lower() == subject.lower())]
            )
            if count > 0:
                available_subjects.append((subject, count))
                print(f"Subject '{subject}' has {count} papers available")
            else:
                print(f"Subject '{subject}' has no papers available")

        return available_subjects
    except Exception as e:
        print(f"Error checking subject availability: {e}")
        return []


def process_subject(subject, args):
    """Process a single subject by calling the extract_biorxiv_tasks.py script.

    Args:
        subject: Subject to process
        args: Command line arguments

    """
    # Clean subject name for command line
    subject.replace(" ", "\\ ")

    # Build command
    cmd = [
        "python",
        "BioAgentOS/bioagentos/scripts/extract_biorxiv_tasks.py",
        "--subject",
        f'"{subject}"',
        "--limit",
        str(args.papers_per_subject),
        "--metadata-path",
        args.metadata_path,
        "--model",
        args.model,
        "--max-paper-length",
        str(args.max_paper_length),
        "--chunk-size",
        str(args.chunk_size),
        "--chunk-overlap",
        str(args.chunk_overlap),
    ]

    # Add optional arguments
    if args.save_pdfs:
        cmd.append("--save-pdfs")
    if args.random_sample:
        cmd.append("--random-sample")

    # Convert command list to string
    cmd_str = " ".join(cmd)

    print(f"\n{'=' * 80}\nProcessing subject: {subject}\n{'=' * 80}")
    print(f"Running command: {cmd_str}")

    # Execute command
    try:
        subprocess.run(cmd_str, shell=True, check=True)
        print(f"Successfully processed subject: {subject}")
    except subprocess.CalledProcessError as e:
        print(f"Error processing subject '{subject}': {e}")


def combine_summaries(base_dir, subjects, papers_per_subject):
    """Combine frequency summaries from all subjects into a single summary.

    Args:
        base_dir: Base directory containing subject results
        subjects: List of subjects that were processed
        papers_per_subject: Number of papers processed per subject

    """
    all_tasks = {}
    all_databases = {}
    all_software = {}

    for subject, _ in subjects:
        # Clean subject name for directory name
        clean_subject = subject.lower().replace(" ", "_").replace("/", "_")
        subject_dir = f"biorxiv_results_{clean_subject}_{papers_per_subject}"
        summary_path = os.path.join(subject_dir, "frequency_summary.json")

        try:
            if os.path.exists(summary_path):
                with open(summary_path) as f:
                    summary = json.load(f)

                    # Merge tasks
                    for task, count in summary.get("tasks", {}).items():
                        if task in all_tasks:
                            all_tasks[task] += count
                        else:
                            all_tasks[task] = count

                    # Merge databases
                    for db, count in summary.get("databases", {}).items():
                        if db in all_databases:
                            all_databases[db] += count
                        else:
                            all_databases[db] = count

                    # Merge software
                    for sw, count in summary.get("software", {}).items():
                        if sw in all_software:
                            all_software[sw] += count
                        else:
                            all_software[sw] = count
            else:
                print(f"Warning: Summary file not found for subject '{subject}'")
        except Exception as e:
            print(f"Error processing summary for subject '{subject}': {e}")

    # Sort by frequency (descending)
    all_tasks = dict(sorted(all_tasks.items(), key=lambda item: item[1], reverse=True))
    all_databases = dict(sorted(all_databases.items(), key=lambda item: item[1], reverse=True))
    all_software = dict(sorted(all_software.items(), key=lambda item: item[1], reverse=True))

    # Create combined summary
    combined_summary = {
        "tasks": all_tasks,
        "databases": all_databases,
        "software": all_software,
    }

    # Save combined summary
    combined_path = os.path.join(base_dir, "combined_summary.json")
    os.makedirs(base_dir, exist_ok=True)

    with open(combined_path, "w") as f:
        json.dump(combined_summary, f, indent=2)

    print(f"Combined summary saved to {combined_path}")

    # Create CSV files for each category
    tasks_df = pd.DataFrame(list(all_tasks.items()), columns=["Task", "Frequency"])
    tasks_df.to_csv(os.path.join(base_dir, "tasks_frequency.csv"), index=False)

    db_df = pd.DataFrame(list(all_databases.items()), columns=["Database", "Frequency"])
    db_df.to_csv(os.path.join(base_dir, "databases_frequency.csv"), index=False)

    sw_df = pd.DataFrame(list(all_software.items()), columns=["Software", "Frequency"])
    sw_df.to_csv(os.path.join(base_dir, "software_frequency.csv"), index=False)

    print(f"CSV summaries saved to {base_dir}")


def main():
    """Main function to run the script."""
    args = parse_arguments()

    # Determine which subjects to process
    subjects_to_process = args.subjects if args.subjects else BIORXIV_SUBJECTS

    # Check which subjects have papers available
    available_subjects = check_subject_availability(args.metadata_path, subjects_to_process)

    if not available_subjects:
        print("No subjects with available papers found. Exiting.")
        return

    if not args.summary_only:
        # Process each subject
        for subject, _count in tqdm(available_subjects, desc="Processing subjects"):
            process_subject(subject, args)

            # Add delay between subjects to avoid rate limiting
            if args.delay > 0:
                print(f"Waiting {args.delay} seconds before processing the next subject...")
                time.sleep(args.delay)

    # Combine summaries from all subjects
    combine_summaries(args.output_dir, available_subjects, args.papers_per_subject)

    print("All subjects processed successfully!")


if __name__ == "__main__":
    import json  # Import here to avoid circular import with the main script

    main()
